"""Machine type classes for Table scitype.

Checks for each class are defined in the "check" method, of signature:

Parameters
----------
obj - object to check
return_metadata - bool, optional, default=False
    if False, returns only "valid" return
    if True, returns all three return objects
var_name: str, optional, default="obj" - name of input in error messages

Returns
-------
valid: bool - whether obj is a valid object of mtype/scitype
msg: str - error message if object is not valid, otherwise None
        returned only if return_metadata is True
metadata: dict - metadata about obj if valid, otherwise None
        returned only if return_metadata is True
    fields:
        "is_univariate": bool, True iff table has one variable
        "is_empty": bool, True iff table has no variables or no instances
        "has_nans": bool, True iff the panel contains NaN values
        "n_instances": int, number of instances/rows in the table
        "n_features": int, number of variables in table
        "feature_names": list of int or object, names of variables in table
"""

__author__ = ["fkiraly"]

import numpy as np
import pandas as pd

from skpro.datatypes._common import _req, _ret
from skpro.datatypes._table._base import BaseTable

PRIMITIVE_TYPES = (float, int, str)


class TablePdDataFrame(BaseTable):
    """Data type: pandas.DataFrame based specification of data frame table.

    Parameters are inferred by check.

    Parameters
    ----------
    is_univariate: bool
        True iff table has one variable
    is_empty: bool
        True iff table has no variables or no instances
    has_nans: bool
        True iff the table contains NaN values
    n_instances: int
        number of instances/rows in the table
    n_features: int
        number of variables in table
    feature_names: list of int or object
        names of variables in table
    """

    _tags = {
        "scitype": "Table",
        "name": "pd_DataFrame_Table",  # any string
        "name_python": "table_pd_df",  # lower_snake_case
        "name_aliases": [],
        "python_version": None,
        "python_dependencies": "pandas",
        "capability:multivariate": True,
        "capability:missing_values": True,
        "capability:index": True,
    }

    def _check(self, obj, return_metadata=False, var_name="obj"):
        """Check if obj is of this data type.

        Parameters
        ----------
        obj : any
            Object to check.
        return_metadata : bool, optional (default=False)
            Whether to return metadata.
        var_name : str, optional (default="obj")
            Name of the variable to check, for use in error messages.

        Returns
        -------
        valid : bool
            Whether obj is of this data type.
        msg : str, only returned if return_metadata is True.
            Error message if obj is not of this data type.
        metadata : dict, only returned if return_metadata is True.
            Metadata dictionary.
        """
        return _check_pddataframe_table(obj, return_metadata, var_name)


def _check_pddataframe_table(obj, return_metadata=False, var_name="obj"):
    metadata = dict()

    if not isinstance(obj, pd.DataFrame):
        msg = f"{var_name} must be a pandas.DataFrame, found {type(obj)}"
        return _ret(False, msg, None, return_metadata)

    # we now know obj is a pd.DataFrame
    index = obj.index
    if _req("is_empty", return_metadata):
        metadata["is_empty"] = len(index) < 1 or len(obj.columns) < 1
    if _req("is_univariate", return_metadata):
        metadata["is_univariate"] = len(obj.columns) < 2
    if _req("n_instances", return_metadata):
        metadata["n_instances"] = len(index)
    if _req("has_nans", return_metadata):
        metadata["has_nans"] = obj.isna().values.any()
    if _req("n_features", return_metadata):
        metadata["n_features"] = len(obj.columns)
    if _req("feature_names", return_metadata):
        metadata["feature_names"] = obj.columns.to_list()

    # check that no dtype is object
    if "object" in obj.dtypes.values:
        msg = f"{var_name} should not have column of 'object' dtype"
        return _ret(False, msg, None, return_metadata)

    return _ret(True, None, metadata, return_metadata)


class TablePdSeries(BaseTable):
    """Data type: pandas.Series based specification of data frame table.

    Parameters are inferred by check.

    Parameters
    ----------
    is_univariate: bool
        True iff table has one variable
    is_empty: bool
        True iff table has no variables or no instances
    has_nans: bool
        True iff the table contains NaN values
    n_instances: int
        number of instances/rows in the table
    n_features: int
        number of variables in table
    feature_names: list of int or object
        names of variables in table
    """

    _tags = {
        "scitype": "Table",
        "name": "pd_Series_Table",  # any string
        "name_python": "table_pd_series",  # lower_snake_case
        "name_aliases": [],
        "python_version": None,
        "python_dependencies": "pandas",
        "capability:multivariate": False,
        "capability:missing_values": True,
        "capability:index": True,
    }

    def _check(self, obj, return_metadata=False, var_name="obj"):
        """Check if obj is of this data type.

        Parameters
        ----------
        obj : any
            Object to check.
        return_metadata : bool, optional (default=False)
            Whether to return metadata.
        var_name : str, optional (default="obj")
            Name of the variable to check, for use in error messages.

        Returns
        -------
        valid : bool
            Whether obj is of this data type.
        msg : str, only returned if return_metadata is True.
            Error message if obj is not of this data type.
        metadata : dict, only returned if return_metadata is True.
            Metadata dictionary.
        """
        return _check_pdseries_table(obj, return_metadata, var_name)


def _check_pdseries_table(obj, return_metadata=False, var_name="obj"):
    metadata = dict()

    if not isinstance(obj, pd.Series):
        msg = f"{var_name} must be a pandas.Series, found {type(obj)}"
        return _ret(False, msg, None, return_metadata)

    # we now know obj is a pd.Series
    index = obj.index
    if _req("is_empty", return_metadata):
        metadata["is_empty"] = len(index) < 1
    if _req("is_univariate", return_metadata):
        metadata["is_univariate"] = True
    if _req("n_instances", return_metadata):
        metadata["n_instances"] = len(index)
    if _req("n_features", return_metadata):
        metadata["n_features"] = 1
    if _req("feature_names", return_metadata):
        if not hasattr(obj, "name") or obj.name is None:
            metadata["feature_names"] = [0]
        else:
            metadata["feature_names"] = [obj.name]

    # check that dtype is not object
    if "object" == obj.dtypes:
        msg = f"{var_name} should not be of 'object' dtype"
        return _ret(False, msg, None, return_metadata)

    # check whether index is equally spaced or if there are any nans
    #   compute only if needed
    if _req("has_nans", return_metadata):
        metadata["has_nans"] = obj.isna().values.any()

    return _ret(True, None, metadata, return_metadata)


class TableNp1D(BaseTable):
    """Data type: 1D np.ndarray based specification of data frame table.

    Parameters are inferred by check.

    Parameters
    ----------
    is_univariate: bool
        True iff table has one variable
    is_empty: bool
        True iff table has no variables or no instances
    has_nans: bool
        True iff the table contains NaN values
    n_instances: int
        number of instances/rows in the table
    n_features: int
        number of variables in table
    feature_names: list of int or object
        names of variables in table
    """

    _tags = {
        "scitype": "Table",
        "name": "numpy1D",  # any string
        "name_python": "table_numpy1d",  # lower_snake_case
        "name_aliases": [],
        "python_version": None,
        "python_dependencies": "numpy",
        "capability:multivariate": False,
        "capability:missing_values": True,
        "capability:index": False,
    }

    def _check(self, obj, return_metadata=False, var_name="obj"):
        """Check if obj is of this data type.

        Parameters
        ----------
        obj : any
            Object to check.
        return_metadata : bool, optional (default=False)
            Whether to return metadata.
        var_name : str, optional (default="obj")
            Name of the variable to check, for use in error messages.

        Returns
        -------
        valid : bool
            Whether obj is of this data type.
        msg : str, only returned if return_metadata is True.
            Error message if obj is not of this data type.
        metadata : dict, only returned if return_metadata is True.
            Metadata dictionary.
        """
        return _check_numpy1d_table(obj, return_metadata, var_name)


def _check_numpy1d_table(obj, return_metadata=False, var_name="obj"):
    metadata = dict()

    if not isinstance(obj, np.ndarray):
        msg = f"{var_name} must be a numpy.ndarray, found {type(obj)}"
        return _ret(False, msg, None, return_metadata)

    if len(obj.shape) != 1:
        msg = f"{var_name} must be 1D numpy.ndarray, but found {len(obj.shape)}D"
        return _ret(False, msg, None, return_metadata)

    # we now know obj is a 1D np.ndarray
    if _req("is_empty", return_metadata):
        metadata["is_empty"] = len(obj) < 1
    if _req("n_instances", return_metadata):
        metadata["n_instances"] = len(obj)
    # 1D numpy arrays are considered univariate
    if _req("is_univariate", return_metadata):
        metadata["is_univariate"] = True
    # check whether there any nans; compute only if requested
    if _req("has_nans", return_metadata):
        metadata["has_nans"] = pd.isnull(obj).any()
    # 1D numpy arrays are considered univariate, with one feature named 0 (integer)
    if _req("n_features", return_metadata):
        metadata["n_features"] = 1
    if _req("feature_names", return_metadata):
        metadata["feature_names"] = [0]

    return _ret(True, None, metadata, return_metadata)


class TableNp2D(BaseTable):
    """Data type: 2D np.ndarray based specification of data frame table.

    Parameters are inferred by check.

    Parameters
    ----------
    is_univariate: bool
        True iff table has one variable
    is_empty: bool
        True iff table has no variables or no instances
    has_nans: bool
        True iff the table contains NaN values
    n_instances: int
        number of instances/rows in the table
    n_features: int
        number of variables in table
    feature_names: list of int or object
        names of variables in table
    """

    _tags = {
        "scitype": "Table",
        "name": "numpy2D",  # any string
        "name_python": "table_numpy2d",  # lower_snake_case
        "name_aliases": [],
        "python_version": None,
        "python_dependencies": "numpy",
        "capability:multivariate": True,
        "capability:missing_values": True,
        "capability:index": False,
    }

    def _check(self, obj, return_metadata=False, var_name="obj"):
        """Check if obj is of this data type.

        Parameters
        ----------
        obj : any
            Object to check.
        return_metadata : bool, optional (default=False)
            Whether to return metadata.
        var_name : str, optional (default="obj")
            Name of the variable to check, for use in error messages.

        Returns
        -------
        valid : bool
            Whether obj is of this data type.
        msg : str, only returned if return_metadata is True.
            Error message if obj is not of this data type.
        metadata : dict, only returned if return_metadata is True.
            Metadata dictionary.
        """
        return _check_numpy2d_table(obj, return_metadata, var_name)


def _check_numpy2d_table(obj, return_metadata=False, var_name="obj"):
    metadata = dict()

    if not isinstance(obj, np.ndarray):
        msg = f"{var_name} must be a numpy.ndarray, found {type(obj)}"
        return _ret(False, msg, None, return_metadata)

    if len(obj.shape) != 2:
        msg = f"{var_name} must be 1D or 2D numpy.ndarray, but found {len(obj.shape)}D"
        return _ret(False, msg, None, return_metadata)

    # we now know obj is a 2D np.ndarray
    if _req("is_empty", return_metadata):
        metadata["is_empty"] = len(obj) < 1 or obj.shape[1] < 1
    if _req("is_univariate", return_metadata):
        metadata["is_univariate"] = obj.shape[1] < 2
    if _req("n_instances", return_metadata):
        metadata["n_instances"] = obj.shape[0]
    # check whether there any nans; compute only if requested
    if _req("has_nans", return_metadata):
        metadata["has_nans"] = pd.isnull(obj).any()
    # 1D numpy arrays are considered univariate, with integer feature names
    if _req("n_features", return_metadata):
        metadata["n_features"] = obj.shape[1]
    if _req("feature_names", return_metadata):
        metadata["feature_names"] = list(range(obj.shape[1]))

    return _ret(True, None, metadata, return_metadata)


class TableListOfDict(BaseTable):
    """Data type: list of dict based specification of data frame table.

    Parameters are inferred by check.

    Parameters
    ----------
    is_univariate: bool
        True iff table has one variable
    is_empty: bool
        True iff table has no variables or no instances
    has_nans: bool
        True iff the table contains NaN values
    n_instances: int
        number of instances/rows in the table
    n_features: int
        number of variables in table
    feature_names: list of int or object
        names of variables in table
    """

    _tags = {
        "scitype": "Table",
        "name": "list_of_dict",  # any string
        "name_python": "table_list_of_dict",  # lower_snake_case
        "name_aliases": [],
        "python_version": None,
        "python_dependencies": "numpy",
        "capability:multivariate": True,
        "capability:missing_values": True,
        "capability:index": False,
    }

    def _check(self, obj, return_metadata=False, var_name="obj"):
        """Check if obj is of this data type.

        Parameters
        ----------
        obj : any
            Object to check.
        return_metadata : bool, optional (default=False)
            Whether to return metadata.
        var_name : str, optional (default="obj")
            Name of the variable to check, for use in error messages.

        Returns
        -------
        valid : bool
            Whether obj is of this data type.
        msg : str, only returned if return_metadata is True.
            Error message if obj is not of this data type.
        metadata : dict, only returned if return_metadata is True.
            Metadata dictionary.
        """
        return _check_list_of_dict_table(obj, return_metadata, var_name)


def _check_list_of_dict_table(obj, return_metadata=False, var_name="obj"):
    metadata = dict()

    if not isinstance(obj, list):
        msg = f"{var_name} must be a list of dict, found {type(obj)}"
        return _ret(False, msg, None, return_metadata)

    if not np.all([isinstance(x, dict) for x in obj]):
        msg = (
            f"{var_name} must be a list of dict, but elements at following "
            f"indices are not dict: {np.where(not isinstance(x, dict) for x in obj)}"
        )
        return _ret(False, msg, None, return_metadata)

    for i, d in enumerate(obj):
        for key in d.keys():
            if not isinstance(d[key], PRIMITIVE_TYPES):
                msg = (
                    "all entries must be of primitive type (str, int, float), but "
                    f"found {type(d[key])} at index {i}, key {key}"
                )

    # we now know obj is a list of dict
    # check whether there any nans; compute only if requested
    if _req("is_univariate", return_metadata):
        multivariate_because_one_row = np.any([len(x) > 1 for x in obj])
        if not multivariate_because_one_row:
            all_keys = np.unique([key for d in obj for key in d.keys()])
            multivariate_because_keys_different = len(all_keys) > 1
            multivariate = multivariate_because_keys_different
        else:
            multivariate = multivariate_because_one_row
        metadata["is_univariate"] = not multivariate
    if _req("has_nans", return_metadata):
        metadata["has_nans"] = np.any(
            [pd.isnull(d[key]) for d in obj for key in d.keys()]
        )
    if _req("is_empty", return_metadata):
        metadata["is_empty"] = len(obj) < 1 or np.all([len(x) < 1 for x in obj])
    if _req("n_instances", return_metadata):
        metadata["n_instances"] = len(obj)

    # this can be expensive, so compute only if requested
    if _req("n_features", return_metadata) or _req("feature_names", return_metadata):
        all_keys = np.unique([key for d in obj for key in d.keys()])
        if _req("n_features", return_metadata):
            metadata["n_features"] = len(all_keys)
        if _req("feature_names", return_metadata):
            metadata["feature_names"] = all_keys.tolist()

    return _ret(True, None, metadata, return_metadata)


class TablePolarsEager(BaseTable):
    """Data type: eager polars DataFrame based specification of data frame table.

    Parameters are inferred by check.

    Parameters
    ----------
    is_univariate: bool
        True iff table has one variable
    is_empty: bool
        True iff table has no variables or no instances
    has_nans: bool
        True iff the table contains NaN values
    n_instances: int
        number of instances/rows in the table
    n_features: int
        number of variables in table
    feature_names: list of int or object
        names of variables in table
    """

    _tags = {
        "scitype": "Table",
        "name": "polars_eager_table",  # any string
        "name_python": "table_polars_eager",  # lower_snake_case
        "name_aliases": [],
        "python_version": None,
        "python_dependencies": ["polars", "pyarrow"],
        "capability:multivariate": True,
        "capability:missing_values": True,
        "capability:index": False,
    }

    def _check(self, obj, return_metadata=False, var_name="obj"):
        """Check if obj is of this data type.

        Parameters
        ----------
        obj : any
            Object to check.
        return_metadata : bool, optional (default=False)
            Whether to return metadata.
        var_name : str, optional (default="obj")
            Name of the variable to check, for use in error messages.

        Returns
        -------
        valid : bool
            Whether obj is of this data type.
        msg : str, only returned if return_metadata is True.
            Error message if obj is not of this data type.
        metadata : dict, only returned if return_metadata is True.
            Metadata dictionary.
        """
        from skpro.datatypes._adapter.polars import check_polars_frame

        return check_polars_frame(obj, return_metadata, var_name, lazy=False)


class TablePolarsLazy(BaseTable):
    """Data type: lazy polars DataFrame based specification of data frame table.

    Parameters are inferred by check.

    Parameters
    ----------
    is_univariate: bool
        True iff table has one variable
    is_empty: bool
        True iff table has no variables or no instances
    has_nans: bool
        True iff the table contains NaN values
    n_instances: int
        number of instances/rows in the table
    n_features: int
        number of variables in table
    feature_names: list of int or object
        names of variables in table
    """

    _tags = {
        "scitype": "Table",
        "name": "polars_lazy_table",  # any string
        "name_python": "table_polars_lazy",  # lower_snake_case
        "name_aliases": [],
        "python_version": None,
        "python_dependencies": ["polars", "pyarrow"],
        "capability:multivariate": True,
        "capability:missing_values": True,
        "capability:index": False,
    }

    def _check(self, obj, return_metadata=False, var_name="obj"):
        """Check if obj is of this data type.

        Parameters
        ----------
        obj : any
            Object to check.
        return_metadata : bool, optional (default=False)
            Whether to return metadata.
        var_name : str, optional (default="obj")
            Name of the variable to check, for use in error messages.

        Returns
        -------
        valid : bool
            Whether obj is of this data type.
        msg : str, only returned if return_metadata is True.
            Error message if obj is not of this data type.
        metadata : dict, only returned if return_metadata is True.
            Metadata dictionary.
        """
        from skpro.datatypes._adapter.polars import check_polars_frame

        return check_polars_frame(obj, return_metadata, var_name, lazy=True)



"""Tests for the show_versions utility."""
import pathlib
import uuid

from skbase.utils.dependencies import _check_soft_dependencies

from skpro.utils._maint._show_versions import (
    DEFAULT_DEPS_TO_SHOW,
    _get_deps_info,
    show_versions,
)


def test_show_versions_runs():
    """Test that show_versions runs without exceptions."""
    # only prints, should return None
    assert show_versions() is None


def test_deps_info():
    """Test that _get_deps_info returns package/version dict as per contract."""
    deps_info = _get_deps_info()
    assert isinstance(deps_info, dict)
    assert set(deps_info.keys()) == {"sktime"}

    deps_info_default = _get_deps_info(DEFAULT_DEPS_TO_SHOW)
    assert isinstance(deps_info_default, dict)
    assert set(deps_info_default.keys()) == set(DEFAULT_DEPS_TO_SHOW)

    KEY_ALIAS = {"sklearn": "scikit-learn", "skbase": "scikit-base"}

    for key in DEFAULT_DEPS_TO_SHOW:
        pkg_name = KEY_ALIAS.get(key, key)
        key_is_available = _check_soft_dependencies(pkg_name, severity="none")
        assert (deps_info_default[key] is None) != key_is_available
        if key_is_available:
            assert _check_soft_dependencies(f"{pkg_name}=={deps_info_default[key]}")
        deps_single_key = _get_deps_info([key])
        assert set(deps_single_key.keys()) == {key}


def test_deps_info_deps_missing_package_present_directory():
    """Test that _get_deps_info does not fail if a dependency is missing."""
    dummy_package_name = uuid.uuid4().hex

    dummy_folder_path = pathlib.Path(dummy_package_name)
    dummy_folder_path.mkdir()

    assert _get_deps_info([dummy_package_name]) == {dummy_package_name: None}

    dummy_folder_path.rmdir()



# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
"""Exponential probability distribution."""

__author__ = ["ShreeshaM07"]

import pandas as pd
from scipy.stats import gamma, rv_continuous

from skpro.distributions.adapters.scipy import _ScipyAdapter


class Gamma(_ScipyAdapter):
    r"""Gamma Distribution.

    Most methods wrap ``scipy.stats.gamma``.

    The Gamma Distribution is parameterized by shape :math:`\alpha` and
    rate :math:`\beta`, such that the pdf is

    .. math:: f(x) = \frac{x^{\alpha-1}\exp\left(-\beta x\right) \beta^{\alpha}}{\tau(\alpha)}

    where :math:`\tau(\alpha)` is the Gamma function.
    For all positive integers, :math:`\tau(\alpha) = (\alpha-1)!`.

    Parameters
    ----------
    alpha : float or array of float (1D or 2D)
        It represents the shape parameter.
    beta : float or array of float (1D or 2D)
        It represents the rate parameter which is also
        inverse of the scale parameter.
    index : pd.Index, optional, default = RangeIndex
    columns : pd.Index, optional, default = RangeIndex

    Examples
    --------
    >>> from skpro.distributions.gamma import Gamma

    >>> d = Gamma(beta=[[1, 1], [2, 3], [4, 5]], alpha=2)
    """  # noqa: E501

    _tags = {
        "capabilities:approx": ["energy", "pdfnorm"],
        "capabilities:exact": ["mean", "var", "pdf", "log_pdf", "cdf", "ppf"],
        "distr:measuretype": "continuous",
        "distr:paramtype": "parametric",
        "broadcast_init": "on",
    }

    def __init__(self, alpha, beta, index=None, columns=None):
        self.alpha = alpha
        self.beta = beta

        super().__init__(index=index, columns=columns)

    def _get_scipy_object(self) -> rv_continuous:
        return gamma

    def _get_scipy_param(self):
        alpha = self._bc_params["alpha"]
        beta = self._bc_params["beta"]
        scale = 1 / beta

        return [], {"a": alpha, "scale": scale}

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator."""
        # array case examples
        params1 = {"alpha": [6, 2.5], "beta": [[1, 1], [2, 3], [4, 5]]}
        params2 = {
            "alpha": 2,
            "beta": 3,
            "index": pd.Index([1, 2, 5]),
            "columns": pd.Index(["a", "b"]),
        }
        # scalar case examples
        params3 = {"alpha": 1.5, "beta": 2.1}

        return [params1, params2, params3]



# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
"""Common private utilities for checkers and converters."""

__author__ = ["fkiraly"]


def _metadata_requested(return_metadata):
    """Return whether some metadata has been requested."""
    return not isinstance(return_metadata, bool) or return_metadata


def _ret(valid, msg, metadata, return_metadata):
    """Return switch for checker functions."""
    if _metadata_requested(return_metadata):
        return valid, msg, metadata
    else:
        return valid


def _req(key, return_metadata):
    """Return whether metadata key is requested, boolean."""
    if isinstance(return_metadata, bool):
        return return_metadata
    elif isinstance(return_metadata, str) and not isinstance(key, list):
        return return_metadata == key
    elif isinstance(return_metadata, str) and isinstance(key, list):
        return return_metadata in key
    elif isinstance(return_metadata, list) and not isinstance(key, list):
        return key in return_metadata
    elif isinstance(return_metadata, list) and isinstance(key, list):
        return len(set(key).intersection(return_metadata)) > 0
    else:
        return False


def _wr(d, key, val, return_metadata):
    """Metadata write switch for checker functions."""
    if _req(key, return_metadata):
        d[key] = val

    return d



"""Adapters to sklearn linear regressors with probabilistic components."""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
# based on sktime pipelines

__author__ = ["fkiraly"]

from skpro.regression.adapters.sklearn import SklearnProbaReg
from skpro.regression.base.adapters import _DelegateWithFittedParamForwarding


class ARDRegression(_DelegateWithFittedParamForwarding):
    """ARD regression, direct adapter to sklearn ARDRegression.

    Fit the weights of a regression model, using an ARD prior. The weights of
    the regression model are assumed to be in Gaussian distributions.
    Also estimate the parameters lambda (precisions of the distributions of the
    weights) and alpha (precision of the distribution of the noise).
    The estimation is done by an iterative procedures (Evidence Maximization)

    Parameters
    ----------
    max_iter : int, default=300
        Maximum number of iterations.

    tol : float, default=1e-3
        Stop the algorithm if w has converged.

    alpha_1 : float, default=1e-6
        Hyper-parameter : shape parameter for the Gamma distribution prior
        over the alpha parameter.

    alpha_2 : float, default=1e-6
        Hyper-parameter : inverse scale parameter (rate parameter) for the
        Gamma distribution prior over the alpha parameter.

    lambda_1 : float, default=1e-6
        Hyper-parameter : shape parameter for the Gamma distribution prior
        over the lambda parameter.

    lambda_2 : float, default=1e-6
        Hyper-parameter : inverse scale parameter (rate parameter) for the
        Gamma distribution prior over the lambda parameter.

    compute_score : bool, default=False
        If True, compute the objective function at each step of the model.

    threshold_lambda : float, default=10 000
        Threshold for removing (pruning) weights with high precision from
        the computation.

    fit_intercept : bool, default=True
        Whether to calculate the intercept for this model. If set
        to false, no intercept will be used in calculations
        (i.e. data is expected to be centered).

    copy_X : bool, default=True
        If True, X will be copied; else, it may be overwritten.

    verbose : bool, default=False
        Verbose mode when fitting the model.

    Attributes
    ----------
    coef_ : array-like of shape (n_features,)
        Coefficients of the regression model (mean of distribution)

    alpha_ : float
       estimated precision of the noise.

    lambda_ : array-like of shape (n_features,)
       estimated precisions of the weights.

    sigma_ : array-like of shape (n_features, n_features)
        estimated variance-covariance matrix of the weights

    scores_ : float
        if computed, value of the objective function (to be maximized)

    n_iter_ : int
        The actual number of iterations to reach the stopping criterion.

    intercept_ : float
        Independent term in decision function. Set to 0.0 if
        ``fit_intercept = False``.

    X_offset_ : float
        If `fit_intercept=True`, offset subtracted for centering data to a
        zero mean. Set to np.zeros(n_features) otherwise.
    """

    def __init__(
        self,
        max_iter=300,
        tol=1e-3,
        alpha_1=1e-6,
        alpha_2=1e-6,
        lambda_1=1e-6,
        lambda_2=1e-6,
        compute_score=False,
        threshold_lambda=10000.0,
        fit_intercept=True,
        copy_X=True,
        verbose=False,
    ):
        self.max_iter = max_iter
        self.tol = tol
        self.alpha_1 = alpha_1
        self.alpha_2 = alpha_2
        self.lambda_1 = lambda_1
        self.lambda_2 = lambda_2
        self.compute_score = compute_score
        self.threshold_lambda = threshold_lambda
        self.fit_intercept = fit_intercept
        self.copy_X = copy_X
        self.verbose = verbose

        from sklearn.linear_model import ARDRegression

        skl_estimator = ARDRegression(
            max_iter=max_iter,
            tol=tol,
            alpha_1=alpha_1,
            alpha_2=alpha_2,
            lambda_1=lambda_1,
            lambda_2=lambda_2,
            compute_score=compute_score,
            threshold_lambda=threshold_lambda,
            fit_intercept=fit_intercept,
            copy_X=copy_X,
            verbose=verbose,
        )

        skpro_est = SklearnProbaReg(skl_estimator, inner_type="np.ndarray")
        self._estimator = skpro_est.clone()

        super().__init__()

    FITTED_PARAMS_TO_FORWARD = [
        "coef_",
        "alpha_",
        "lambda_",
        "sigma_",
        "scores_",
        "n_iter_",
        "intercept_",
        "X_offset_",
    ]

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return `"default"` set.

        Returns
        -------
        params : dict or list of dict, default = {}
            Parameters to create testing instances of the class
            Each dict are parameters to construct an "interesting" test instance, i.e.,
            `MyClass(**params)` or `MyClass(**params[i])` creates a valid test instance.
            `create_test_instance` uses the first (or only) dictionary in `params`
        """
        param1 = {}
        param2 = {
            "max_iter": 300,
            "tol": 2e-3,
            "alpha_1": 2e-6,
            "alpha_2": 2e-6,
            "lambda_1": 2e-6,
            "lambda_2": 2e-6,
            "compute_score": True,
            "threshold_lambda": 15000.0,
            "fit_intercept": False,
        }
        return [param1, param2]


class BayesianRidge(_DelegateWithFittedParamForwarding):
    """Bayesian ridge regression, direct adapter to sklearn BayesianRidge.

    Fit a Bayesian ridge model. See the Notes section for details on this
    implementation and the optimization of the regularization parameters
    lambda (precision of the weights) and alpha (precision of the noise).

    Parameters
    ----------
    max_iter : int, default=300
        Maximum number of iterations over the complete dataset before
        stopping independently of any early stopping criterion.

    tol : float, default=1e-3
        Stop the algorithm if w has converged.

    alpha_1 : float, default=1e-6
        Hyper-parameter : shape parameter for the Gamma distribution prior
        over the alpha parameter.

    alpha_2 : float, default=1e-6
        Hyper-parameter : inverse scale parameter (rate parameter) for the
        Gamma distribution prior over the alpha parameter.

    lambda_1 : float, default=1e-6
        Hyper-parameter : shape parameter for the Gamma distribution prior
        over the lambda parameter.

    lambda_2 : float, default=1e-6
        Hyper-parameter : inverse scale parameter (rate parameter) for the
        Gamma distribution prior over the lambda parameter.

    alpha_init : float, default=None
        Initial value for alpha (precision of the noise).
        If not set, alpha_init is 1/Var(y).

    lambda_init : float, default=None
        Initial value for lambda (precision of the weights).
        If not set, lambda_init is 1.

    compute_score : bool, default=False
        If True, compute the log marginal likelihood at each iteration of the
        optimization.

    fit_intercept : bool, default=True
        Whether to calculate the intercept for this model.
        The intercept is not treated as a probabilistic parameter
        and thus has no associated variance. If set
        to False, no intercept will be used in calculations
        (i.e. data is expected to be centered).

    copy_X : bool, default=True
        If True, X will be copied; else, it may be overwritten.

    verbose : bool, default=False
        Verbose mode when fitting the model.

    Attributes
    ----------
    coef_ : array-like of shape (n_features,)
        Coefficients of the regression model (mean of distribution)

    intercept_ : float
        Independent term in decision function. Set to 0.0 if
        `fit_intercept = False`.

    alpha_ : float
       Estimated precision of the noise.

    lambda_ : float
       Estimated precision of the weights.

    sigma_ : array-like of shape (n_features, n_features)
        Estimated variance-covariance matrix of the weights

    scores_ : array-like of shape (n_iter_+1,)
        If computed_score is True, value of the log marginal likelihood (to be
        maximized) at each iteration of the optimization. The array starts
        with the value of the log marginal likelihood obtained for the initial
        values of alpha and lambda and ends with the value obtained for the
        estimated alpha and lambda.

    n_iter_ : int
        The actual number of iterations to reach the stopping criterion.

    X_offset_ : ndarray of shape (n_features,)
        If `fit_intercept=True`, offset subtracted for centering data to a
        zero mean. Set to np.zeros(n_features) otherwise.
    """

    def __init__(
        self,
        max_iter=300,
        tol=1e-3,
        alpha_1=1e-6,
        alpha_2=1e-6,
        lambda_1=1e-6,
        lambda_2=1e-6,
        alpha_init=None,
        lambda_init=None,
        compute_score=False,
        fit_intercept=True,
        copy_X=True,
        verbose=False,
    ):
        self.max_iter = max_iter
        self.tol = tol
        self.alpha_1 = alpha_1
        self.alpha_2 = alpha_2
        self.lambda_1 = lambda_1
        self.lambda_2 = lambda_2
        self.alpha_init = alpha_init
        self.lambda_init = lambda_init
        self.compute_score = compute_score
        self.fit_intercept = fit_intercept
        self.copy_X = copy_X
        self.verbose = verbose

        from sklearn.linear_model import BayesianRidge

        skl_estimator = BayesianRidge(
            max_iter=max_iter,
            tol=tol,
            alpha_1=alpha_1,
            alpha_2=alpha_2,
            lambda_1=lambda_1,
            lambda_2=lambda_2,
            alpha_init=alpha_init,
            lambda_init=lambda_init,
            compute_score=compute_score,
            fit_intercept=fit_intercept,
            copy_X=copy_X,
            verbose=verbose,
        )

        skpro_est = SklearnProbaReg(skl_estimator)
        self._estimator = skpro_est.clone()

        super().__init__()

    FITTED_PARAMS_TO_FORWARD = [
        "coef_",
        "alpha_",
        "lambda_",
        "sigma_",
        "scores_",
        "n_iter_",
        "intercept_",
        "X_offset_",
    ]

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return `"default"` set.

        Returns
        -------
        params : dict or list of dict, default = {}
            Parameters to create testing instances of the class
            Each dict are parameters to construct an "interesting" test instance, i.e.,
            `MyClass(**params)` or `MyClass(**params[i])` creates a valid test instance.
            `create_test_instance` uses the first (or only) dictionary in `params`
        """
        param1 = {}
        param2 = {
            "max_iter": 300,
            "tol": 2e-3,
            "alpha_1": 2e-6,
            "alpha_2": 2e-6,
            "lambda_1": 2e-6,
            "lambda_2": 2e-6,
            "compute_score": True,
            "fit_intercept": False,
        }
        return [param1, param2]



"""Residual regression - one regressor for mean, one for scale."""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)

__author__ = ["fkiraly"]

import numpy as np
import pandas as pd
from sklearn import clone

from skpro.regression.base import BaseProbaRegressor
from skpro.utils.numpy import flatten_to_1D_if_colvector
from skpro.utils.sklearn import prep_skl_df


class ResidualDouble(BaseProbaRegressor):
    """Residual double regressor.

    Make a parametric probabilistic prediction using two tabular regressors, with
    one tabular regressor predicting the mean, and one the deviation from the mean.

    The mean is predicted by ``estimator``. The residual is predicted by
    ``estimator_resid``. The residual is transformed by ``residual_trafo``.
    The predicted mean and residual are passed to a distribution specified by
    ``distr_type``, and possibly ``distr_params``, ``distr_loc_scale_name``.

    The residuals predicted on the training data are used to fit
    ``estimator_resid``. If ``cv`` is passed, the residuals are out-of-sample
    according to ``cv``, otherwise in-sample.

    ``use_y_pred`` determines whether the predicted mean is used as a feature
    in predicting the residual.

    A formal description of the algorithm follows.

    In ``fit``, given training data ``X``, ``y``:

    1. Fit clone ``estimator_`` of ``estimator`` to predict ``y`` from ``X``,
       i.e., ``fit(X, y)``.
    2. Predict mean label ``y_pred`` for ``X`` using a clone of ``estimator``.
       If ``cv`` is ``None``, this is via plain ``estimator.predict(X)``.
       If ``cv`` is not ``None``, out-of-sample predictions are obtained via ``cv``.
       In this case, indices not appearing in ``cv`` are predicted in-sample.
    3. Compute residual ``resid`` as ``residual_trafo(y - y_pred)``.
       If ``residual_trafo`` is a transformer, ``residual_trafo.fit_transform`` is used.
    4. Fit clone ``estimator_resid_`` of ``estimator_resid``
       to predict ``resid`` from ``X``, i.e., ``fit(X, resid)``.
       If ``use_y_pred`` is ``True``, ``y_pred`` is used as a feature in predicting.

    In ``predict``, given test data ``X``:

    1. Predict mean label ``y_pred`` for ``X`` using ``estimator_.predict(X)``.
    2. Return ``y_pred``.

    In ``predict_proba``, given test data ``X``:

    1. Predict mean label ``y_pred`` for ``X`` using ``estimator_.predict(X)``.
    2. Predict residual ``resid`` for ``X`` using ``estimator_resid_.predict(X)``.
       If ``use_y_pred`` is ``True``, ``y_pred`` is used as a feature in predicting.
    3. Predict distribution ``y_pred_proba`` for ``X`` as follows:
       The location parameter is ``y_pred``. The scale parameter is ``resid``.
       Further parameters can be specified via ``distr_params``.
    4. Return ``y_pred_proba``.

    Parameters
    ----------
    estimator : sklearn regressor
        estimator predicting the mean or location
    estimator_resid : sklearn regressor
        estimator predicting the scale of the residual
        default = sklearn DummyRegressor(strategy="mean")
    residual_trafo : str, or transformer, default="absolute"
        determines the labels predicted by ``estimator_resid``
        absolute = absolute residuals
        squared = squared residuals
        if transformer, applies fit_transform to batch of signed residuals
    distr_type : str or BaseDistribution, default = "Normal"
        type of distribution to predict
        str options are "Normal", "Laplace", "Cauchy", "t"
    distr_loc_scale_name : tuple of length two, default = ("loc", "scale")
        names of the parameters in the distribution to use for location and scale
        if ``distr_type`` is a string, this is overridden to the correct parameters
        if ``distr_type`` is a BaseDistribution, this is used to determine the
        location and scale parameters that the predictions are passed to
    distr_params : dict, default = {}
        parameters to pass to the distribution
        must be valid parameters of ``distr_type``, if ``BaseDistribution``
        must be default or dict with key ``df``, if ``t`` distribution
    use_y_pred : bool, default=False
        whether to use the predicted location in predicting the scale of the residual
    cv : optional, sklearn cv splitter, default = None
        if passed, will be used to obtain out-of-sample residuals according to cv
        instead of in-sample residuals in ``fit`` of this estimator
    min_scale : float, default=1e-10
        minimum scale parameter. If smaller scale parameter is predicted by
        ``estimator_resid``, will be clipped to this value

    Attributes
    ----------
    estimator_ : sklearn regressor, clone of ``estimator``
        fitted estimator predicting the mean or location
    estimator_resid_ : sklearn regressor, clone of ``estimator_resid``
        fitted estimator predicting the scale of the residual

    Examples
    --------
    >>> from skpro.regression.residual import ResidualDouble
    >>> from sklearn.ensemble import RandomForestRegressor
    >>> from sklearn.linear_model import LinearRegression
    >>> from sklearn.datasets import load_diabetes
    >>>
    >>> X, y = load_diabetes(return_X_y=True, as_frame=True)
    >>> reg_mean = LinearRegression()
    >>> reg_resid = RandomForestRegressor()
    >>> reg_proba = ResidualDouble(reg_mean, reg_resid)
    >>>
    >>> reg_proba.fit(X, y)
    ResidualDouble(...)
    >>> y_pred_mean = reg_proba.predict(X)
    >>> y_pred_proba = reg_proba.predict_proba(X)
    """

    _tags = {"capability:missing": True}

    def __init__(
        self,
        estimator,
        estimator_resid=None,
        residual_trafo="absolute",
        distr_type="Normal",
        distr_loc_scale_name=None,
        distr_params=None,
        use_y_pred=False,
        cv=None,
        min_scale=1e-10,
    ):
        self.estimator = estimator
        self.estimator_resid = estimator_resid
        self.residual_trafo = residual_trafo
        self.distr_type = distr_type
        self.distr_loc_scale_name = distr_loc_scale_name
        self.distr_params = distr_params
        self.use_y_pred = use_y_pred
        self.cv = cv
        self.min_scale = min_scale

        super().__init__()

        self.estimator_ = clone(estimator)

        if estimator_resid is None:
            from sklearn.dummy import DummyRegressor

            self.estimator_resid_ = DummyRegressor(strategy="mean")
        else:
            self.estimator_resid_ = clone(estimator_resid)

    def _predict_residuals_cv(self, X, y, cv, est):
        """Predict out-of-sample residuals for y from X using cv.

        Parameters
        ----------
        X : pandas DataFrame
            feature instances to fit regressor to
        y : pandas DataFrame, must be same length as X
            labels to fit regressor to
        cv : sklearn cv splitter
            cv splitter to use for out-of-sample predictions

        Returns
        -------
        y_pred : pandas DataFrame, same length as `X`, same columns as `y` in `fit`
            labels predicted for `X`
        """
        est = self.estimator_resid_
        method = "predict"
        y_pred = y.copy()

        for tr_idx, tt_idx in cv.split(X):
            X_train = X.iloc[tr_idx]
            X_test = X.iloc[tt_idx]
            y_train = y[tr_idx]
            fitted_est = clone(est).fit(X_train, y_train)
            y_pred[tt_idx] = getattr(fitted_est, method)(X_test)

        return y_pred

    def _fit(self, X, y):
        """Fit regressor to training data.

        Writes to self:
            Sets fitted model attributes ending in "_".

        Parameters
        ----------
        X : pandas DataFrame
            feature instances to fit regressor to
        y : pandas DataFrame, must be same length as X
            labels to fit regressor to

        Returns
        -------
        self : reference to self
        """
        est = self.estimator_
        est_r = self.estimator_resid_
        residual_trafo = self.residual_trafo
        cv = self.cv
        use_y_pred = self.use_y_pred

        self._y_cols = y.columns

        # coerce X to pandas DataFrame with string column names
        X = prep_skl_df(X, copy_df=True)

        # flatten column vector to 1D array to avoid sklearn complaints
        y = y.values
        y = flatten_to_1D_if_colvector(y)

        est.fit(X, y)

        if cv is None:
            y_pred = est.predict(X)
        else:
            y_pred = self._predict_residuals_cv(X, y, cv, est)

        if residual_trafo == "absolute":
            resids = np.abs(y - y_pred)
        elif residual_trafo == "squared":
            resids = (y - y_pred) ** 2
        else:
            resids = residual_trafo.fit_transform(y - y_pred)

        resids = flatten_to_1D_if_colvector(resids)

        if use_y_pred:
            y_ix = {"index": X.index, "columns": self._y_cols}
            X_r = pd.concat([X, pd.DataFrame(y_pred, **y_ix)], axis=1)
        else:
            X_r = X

        # coerce X to pandas DataFrame with string column names
        X_r = prep_skl_df(X_r, copy_df=True)

        est_r.fit(X_r, resids)

        return self

    def _predict(self, X):
        """Predict labels for data from features.

        State required:
            Requires state to be "fitted" = self.is_fitted=True

        Accesses in self:
            Fitted model attributes ending in "_"

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for

        Returns
        -------
        y : pandas DataFrame, same length as `X`, same columns as `y` in `fit`
            labels predicted for `X`
        """
        est = self.estimator_

        X = prep_skl_df(X, copy_df=True)

        y_pred = est.predict(X)
        y_pred = pd.DataFrame(y_pred, columns=self._y_cols, index=X.index)

        return y_pred

    def _predict_proba(self, X):
        """Predict distribution over labels for data from features.

        State required:
            Requires state to be "fitted".

        Accesses in self:
            Fitted model attributes ending in "_"

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for

        Returns
        -------
        y : skpro BaseDistribution, same length as `X`
            labels predicted for `X`
        """
        est = self.estimator_
        est_r = self.estimator_resid_
        use_y_pred = self.use_y_pred
        distr_type = self.distr_type
        distr_loc_scale_name = self.distr_loc_scale_name
        distr_params = self.distr_params
        min_scale = self.min_scale

        # coerce X to pandas DataFrame with string column names
        X = prep_skl_df(X, copy_df=True)

        n_cols = len(self._y_cols)

        if distr_params is None:
            distr_params = {}

        # predict location - this is the same as in _predict
        y_pred_loc = est.predict(X)
        y_pred_loc = y_pred_loc.reshape(-1, n_cols)

        # predict scale
        # if use_y_pred, use predicted location as feature
        if use_y_pred:
            y_ix = {"index": X.index, "columns": self._y_cols}
            X_r = pd.concat([X, pd.DataFrame(y_pred_loc, **y_ix)], axis=1)
        # if not use_y_pred, use only original features
        else:
            X_r = X

        # coerce X to pandas DataFrame with string column names
        X_r = prep_skl_df(X_r, copy_df=True)

        y_pred_scale = est_r.predict(X_r)
        y_pred_scale = y_pred_scale.clip(min=min_scale)
        y_pred_scale = y_pred_scale.reshape(-1, n_cols)

        # create distribution with predicted scale and location
        # we deal with string distr_types by getting class and param names
        if distr_type == "Normal":
            from skpro.distributions.normal import Normal

            distr_type = Normal
            distr_loc_scale_name = ("mu", "sigma")
        elif distr_type == "Laplace":
            from skpro.distributions.laplace import Laplace

            distr_type = Laplace
            distr_loc_scale_name = ("mu", "scale")
        elif distr_type in ["Cauchy", "t"]:
            from skpro.distributions.t import TDistribution

            distr_type = TDistribution
            distr_loc_scale_name = ("mu", "sigma")

        # collate all parameters for the distribution constructor
        # distribution params, if passed
        params = distr_params
        # row/column index
        ix = {"index": X.index, "columns": self._y_cols}
        params.update(ix)
        # location and scale
        loc_scale = {
            distr_loc_scale_name[0]: y_pred_loc,
            distr_loc_scale_name[1]: y_pred_scale,
        }
        params.update(loc_scale)

        # create distribution and return
        y_pred = distr_type(**params)
        return y_pred

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return `"default"` set.

        Returns
        -------
        params : dict or list of dict, default = {}
            Parameters to create testing instances of the class
            Each dict are parameters to construct an "interesting" test instance, i.e.,
            `MyClass(**params)` or `MyClass(**params[i])` creates a valid test instance.
            `create_test_instance` uses the first (or only) dictionary in `params`
        """
        from sklearn.ensemble import RandomForestRegressor
        from sklearn.linear_model import LinearRegression
        from sklearn.model_selection import KFold

        params1 = {"estimator": RandomForestRegressor()}
        params2 = {
            "estimator": LinearRegression(),
            "estimator_resid": RandomForestRegressor(),
            "min_scale": 1e-7,
            "residual_trafo": "squared",
            "use_y_pred": True,
            "distr_type": "Laplace",
        }
        params3 = {
            "estimator": LinearRegression(),
            "estimator_resid": RandomForestRegressor(),
            "min_scale": 1e-6,
            "use_y_pred": True,
            "distr_type": "t",
            "distr_params": {"df": 3},
            "cv": KFold(n_splits=3),
        }
        params4 = {"estimator": RandomForestRegressor(), "cv": KFold(n_splits=3)}

        return [params1, params2, params3, params4]



# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
"""Logistic probability distribution."""

__author__ = ["malikrafsan"]

import numpy as np
import pandas as pd

from skpro.distributions.base import BaseDistribution


class Logistic(BaseDistribution):
    r"""Logistic distribution.

    The logistic distribution is parametrized by a mean parameter :math:`\mu`,
    and scale parameter :math:`s`, such that the cdf is given by:

    .. math:: F(x) = \frac{1}{1 + \exp\left(\frac{x - \mu}{s}\right)}

    The scale parameter :math:`s` is represented by the parameter ``scale``,
    and the mean parameter :math:`\mu` by the parameter ``mu``.

    Parameters
    ----------
    mu : float or array of float (1D or 2D)
        mean of the logistic distribution
    scale : float or array of float (1D or 2D), must be positive
        scale parameter of the distribution
    index : pd.Index, optional, default = RangeIndex
    columns : pd.Index, optional, default = RangeIndex

    Examples
    --------
    >>> from skpro.distributions.logistic import Logistic

    >>> l = Logistic(mu=[[0, 1], [2, 3], [4, 5]], scale=1)
    """

    _tags = {
        "capabilities:approx": ["pdfnorm", "energy"],
        "capabilities:exact": ["mean", "var", "pdf", "log_pdf", "cdf", "ppf"],
        "distr:measuretype": "continuous",
        "distr:paramtype": "parametric",
        "broadcast_init": "on",
    }

    def __init__(self, mu, scale, index=None, columns=None):
        self.mu = mu
        self.scale = scale

        super().__init__(index=index, columns=columns)

    def _mean(self):
        """Return expected value of the distribution.

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            expected value of distribution (entry-wise)
        """
        return self._bc_params["mu"]

    def _var(self):
        r"""Return variance of the distribution.

        Let :math:`X` be a random variable with the distribution of `self`.
        Returns the variance :math:`\mathbb{V}[X] = \frac{\mathbb{S}^2 \times \pi^3}{3}`

        Returns
        -------
        pd.DataFrame with same rows, columns as `self`
        variance of distribution (entry-wise)
        """
        scale = self._bc_params["scale"]
        var_arr = (scale**2 * np.pi**2) / 3
        return var_arr

    def _pdf(self, x):
        """Probability density function.

        Parameters
        ----------
        x : 2D np.ndarray, same shape as ``self``
            values to evaluate the pdf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            pdf values at the given points
        """
        mu = self._bc_params["mu"]
        scale = self._bc_params["scale"]

        numerator = np.exp(-(x - mu) / scale)
        denominator = scale * ((1 + np.exp(-(x - mu) / scale)) ** 2)
        pdf_arr = numerator / denominator
        return pdf_arr

    def _log_pdf(self, x):
        """Logarithmic probability density function.

        Parameters
        ----------
        x : 2D np.ndarray, same shape as ``self``
            values to evaluate the pdf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            log pdf values at the given points
        """
        mu = self._bc_params["mu"]
        scale = self._bc_params["scale"]

        y = -(x - mu) / scale
        lpdf_arr = y - np.log(scale) - 2 * np.logaddexp(0, y)
        return lpdf_arr

    def _cdf(self, x):
        """Cumulative distribution function.

        Parameters
        ----------
        x : 2D np.ndarray, same shape as ``self``
            values to evaluate the cdf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            cdf values at the given points
        """
        mu = self._bc_params["mu"]
        scale = self._bc_params["scale"]

        cdf_arr = (1 + np.tanh((x - mu) / (2 * scale))) / 2
        return cdf_arr

    def _ppf(self, p):
        """Quantile function = percent point function = inverse cdf.

        Parameters
        ----------
        p : 2D np.ndarray, same shape as ``self``
            values to evaluate the ppf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            ppf values at the given points
        """
        mu = self._bc_params["mu"]
        scale = self._bc_params["scale"]

        ppf_arr = mu + scale * np.log(p / (1 - p))
        return ppf_arr

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator."""
        # array case examples
        params1 = {"mu": [[0, 1], [2, 3], [4, 5]], "scale": 1}
        params2 = {
            "mu": 0,
            "scale": 1,
            "index": pd.Index([1, 2, 5]),
            "columns": pd.Index(["a", "b"]),
        }
        # scalar case examples
        params3 = {"mu": -2, "scale": 2}

        return [params1, params2, params3]



"""Registry and lookup functionality."""

from skpro.registry._lookup import all_objects, all_tags
from skpro.registry._scitype import scitype
from skpro.registry._tags import (
    OBJECT_TAG_LIST,
    OBJECT_TAG_REGISTER,
    check_tag_is_valid,
)

__all__ = [
    "OBJECT_TAG_LIST",
    "OBJECT_TAG_REGISTER",
    "all_objects",
    "all_tags",
    "check_tag_is_valid",
    "scitype",
]



"""Survival Process Logarithmic Loss for distributional predictions."""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)

import numpy as np
import pandas as pd

from skpro.metrics.base import BaseSurvDistrMetric


class SPLL(BaseSurvDistrMetric):
    r"""Survival Process Logarithmic Loss for distributional predictions.

    Same as the negative log-likelihood of the survival process (see [1]_),
    and therefore a proper scoring rule for survival predictions.

    For a predictive distribution :math:`d` with pdf :math:`d.p`,
    survival function :math:`d.S`, a ground truth value :math:`y`
    and censoring indicator :math:`\Delta`, taking values 1 (censored) and
    0 (uncensored), the survival process logarithmic loss is defined as
    :math:`L((y, \Delta), d) := (\Delta - 1) (\log d.p(y)) - \Delta \log d.S(y)`.
    Logarithms are natural logarithms.

    Where :math:`d.S = 1 - d.F` for the cdf :math:`d.F` of :math:`d`.

    To obtain the loss from formula 7.1.2 in [1]_, condition on
    :math:`N(A) \le 1` (i.e., no more than one event in the interval :math:`A`),
    use that :math:`\Delta = 1` iff :math:`N(A) = 0`,
    and :math:`-\log d.S(y) = \int_A d.\lambda(t) \; dt`, with
    :math:`d.\lambda = -d.p / d.S`.

    `evaluate` computes the average test sample loss.
    `evaluate_by_index` produces the loss sample by test data point.
    `multivariate` controls averaging over variables.

    Parameters
    ----------
    multioutput : {'raw_values', 'uniform_average'} or array-like of shape \
            (n_outputs,), default='uniform_average'
        Defines whether and how to aggregate metric for across variables.
        If 'uniform_average' (default), errors are mean-averaged across variables.
        If array-like, errors are weighted averaged across variables, values as weights.
        If 'raw_values', does not average errors across variables, columns are retained.
    multivariate : bool, optional, default=False
        if True, behaves as multivariate log-loss
        log-loss is computed for entire row, results one score per row
        if False, is univariate log-loss
        log-loss is computed per variable marginal, results in many scores per row

    References
    ----------
    .. [1] Daley DJ, Vere-Jones. An Introduction to the Theory of Point Processes,
        2nd Edition, 2003, Springer, New York. Formula 7.1.2.
    """

    _tags = {
        "authors": "fkiraly",
        "capability:survival": True,
        "scitype:y_pred": "pred_proba",
        "lower_is_better": True,
    }

    def __init__(self, multioutput="uniform_average", multivariate=False):
        self.multivariate = multivariate
        super().__init__(multioutput=multioutput)

    def _evaluate_by_index(self, y_true, y_pred, **kwargs):
        C_true = kwargs.get("C_true", None)

        if C_true is None:
            # then all uncensored, Delta = 0
            res = -y_pred.log_pdf(y_true)
        else:
            cont_term = -y_pred.log_pdf(y_true) * (1 - C_true.to_numpy())
            disc_term = np.log(1 - y_pred.cdf(y_true)) * C_true.to_numpy()
            res = cont_term + disc_term

        if self.multivariate:
            return pd.DataFrame(res.mean(axis=1), columns=["SPLL"])
        else:
            return res

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Retrieve test parameters."""
        params1 = {}
        params2 = {"multivariate": True}
        return [params1, params2]



# LEGACY MODULE - TODO: remove or refactor


def not_existing(f):
    """
    Decorates an interface method to declare it theoretically non existent

    Parameters
    ----------
    f   Method to decorate

    Returns
    -------
    Decorated method
    """
    f.not_existing = True

    return f


def ensure_existence(f):
    """Ensures that method is not marked as non_existent

    Parameters
    ----------
    f  Method

    Raises
    ------
    NotImplementedError if the method is marked as non existent

    Returns
    -------
    Method f
    """
    if getattr(f, "not_existing", False):
        raise NotImplementedError(
            "The distribution has no " + f.__name__ + " function. "
            "You may use an adapter that supports its approximation."
        )

    return f


def to_percent(value, return_float=True):
    """Converts values into a percent representation

    Args:
        value: int/float
            Number representing a percentage
        return_float: bool
            If true, float representing the percentage is returned

    Returns: int/float
        A percentage
    """

    def percent(p):
        if return_float:
            return float(p)
        else:
            return int(round(p * 100))

    if isinstance(value, int):
        value = float(value) / 100.0

    if value <= 0:
        return percent(0)
    else:
        return percent(value)



"""Adapters to sklearn linear regressors with probabilistic components."""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
# based on sktime pipelines

import pandas as pd

from skpro.regression.base import BaseProbaRegressor
from skpro.utils.sklearn import prep_skl_df


class PoissonRegressor(BaseProbaRegressor):
    """Poisson regression, direct adapter to sklearn PoissonRegressor.

    Generalized Linear Model with a Poisson distribution.
    This regressor uses the 'log' link function.

    Parameters
    ----------
    alpha : float, default=1.0
        Constant that multiplies the penalty term. Defaults to 1.0.
        See the notes for the exact mathematical meaning of this
        parameter. alpha = 0 is equivalent to unpenalized GLMs.

    fit_intercept : bool, default=True
        Whether to fit an intercept term.

    solver : {'lbfgs', 'newton-cholesky'}, default='lbfgs'
        Algorithm to use in the optimization problem.

    'lbfgs' is an optimization algorithm that approximates the BFGS algorithm
    'newton-cholesky' uses a Newton-CG variant of Newton's method.

    max_iter : int, default=100
        The maximal number of iterations for the solver.

    tol : float, default=1e-4
        The convergence tolerance. If it is not None, training will stop
        when (loss > best_loss - tol) for n_iter_no_change consecutive
        epochs.

    verbose : int, default=0
        For the 'sag' and 'lbfgs' solvers set verbose to any positive
        number for verbosity.

    warm_start : bool, default=False
        When set to True, reuse the solution of the previous call to fit as
        initialization, otherwise, just erase the previous solution.

    Attributes
    ----------
    coef_ : array-like of shape (n_features,)
        Coefficients of the regression model (mean of distribution)

    intercept_ : float
        Independent term in decision function.

    n_iter_ : int
        The actual number of iterations before reaching the stopping criterion.

    n_features_in_ : int
        Number of features seen during :term:'fit'.

    feature_names_in_ : ndarray of shape (n_features,)
        Names of features seen during :term:'fit'.
    """

    _tags = {
        "capability:multioutput": False,
        "capability:missing": False,
        "X_inner_mtype": "pd_DataFrame_Table",
        "y_inner_mtype": "pd_DataFrame_Table",
    }

    def __init__(
        self,
        alpha=1.0,
        fit_intercept=True,
        max_iter=100,
        tol=1e-4,
        verbose=0,
        warm_start=False,
    ):
        self.alpha = alpha
        self.fit_intercept = fit_intercept
        self.max_iter = max_iter
        self.tol = tol
        self.verbose = verbose
        self.warm_start = warm_start

        super().__init__()

        from sklearn.linear_model import PoissonRegressor

        skl_estimator = PoissonRegressor(
            alpha=alpha,
            fit_intercept=fit_intercept,
            max_iter=max_iter,
            tol=tol,
            verbose=verbose,
            warm_start=warm_start,
        )

        self.estimator_ = skl_estimator

    FITTED_PARAMS_TO_FORWARD = [
        "coef_",
        "intercept_",
        "n_iter_",
    ]

    def _fit(self, X, y):
        """Fit regressor to training data.

        Writes to self:
            Sets fitted model attributes ending in "_".

        Parameters
        ----------
        X : pandas DataFrame
            feature instances to fit regressor to
        y : pandas DataFrame, must be same length as X
            labels to fit regressor to

        Returns
        -------
        self : reference to self
        """
        X_inner = prep_skl_df(X).to_numpy()
        y_inner = prep_skl_df(y).to_numpy()

        self._y_cols = y.columns

        if len(y_inner.shape) > 1 and y_inner.shape[1] == 1:
            y_inner = y_inner[:, 0]

        estimator = self.estimator_
        estimator.fit(X=X_inner, y=y_inner)

        for attr in self.FITTED_PARAMS_TO_FORWARD:
            setattr(self, attr, getattr(estimator, attr))

        return self

    def _predict(self, X):
        """Predict labels for data from features.

        State required:
            Requires state to be "fitted" = self.is_fitted=True

        Accesses in self:
            Fitted model attributes ending in "_"

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for

        Returns
        -------
        y : pandas DataFrame, same length as `X`, same columns as `y` in `fit`
            labels predicted for `X`
        """
        X_inner = prep_skl_df(X).to_numpy()
        y_pred = self.estimator_.predict(X_inner)
        y_pred_df = pd.DataFrame(y_pred, index=X.index, columns=self._y_cols)
        return y_pred_df

    def _predict_var(self, X):
        """Compute/return variance predictions."""
        return self._predict(X)  # Poisson variance is equal to mean

    def _predict_proba(self, X):
        """Predict distribution over labels for data from features.

        State required:
            Requires state to be "fitted".

        Accesses in self:
            Fitted model attributes ending in "_"

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for

        Returns
        -------
        y_pred : skpro BaseDistribution, same length as `X`
            labels predicted for `X`
        """
        from skpro.distributions.poisson import Poisson

        y_cols = self._y_cols
        y_pred = self.predict(X).values
        y_pred_proba = Poisson(y_pred, index=X.index, columns=y_cols)
        return y_pred_proba

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return `"default"` set.

        Returns
        -------
        params : dict or list of dict, default = {}
            Parameters to create testing instances of the class
            Each dict are parameters to construct an "interesting" test instance, i.e.,
            `MyClass(**params)` or `MyClass(**params[i])` creates a valid test instance.
            `create_test_instance` uses the first (or only) dictionary in `params`
        """
        param1 = {}
        param2 = {
            "alpha": 2.0,
            "fit_intercept": False,
            "max_iter": 200,
            "tol": 2e-4,
            "verbose": 1,
            "warm_start": True,
        }
        return [param1, param2]



"""Interface adapter to lifelines Fisk AFT model."""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)

__author__ = ["fkiraly"]

import numpy as np

from skpro.distributions.fisk import Fisk
from skpro.survival.adapters.lifelines import _LifelinesAdapter
from skpro.survival.base import BaseSurvReg


class AFTFisk(_LifelinesAdapter, BaseSurvReg):
    r"""Log-Logitsic/Fisk AFT model, from lifelines.

    Direct interface to ``lifelines.fitters.LogLogisticAFTFitter``,
    by ``CamDavidsonPilon``.

    This class implements a Log-Logistic AFT model. The model has parametric
    form, with :math:`\alpha(x) = \exp\left(a_0 + a_1x_1 + ... + a_n x_n \right)`,
    and, optionally, :math:`\beta(y) = \exp\left(b_0 + b_1 y_1 + ... + b_m y_m \right)`.

    with predictive distribution being Fisk (aka log-logistic), with
    scale parameter :math:`\alpha(x)` and shape parameter (exponent) :math:`\beta(y)`.

    The :math:`\alpha` (scale) parameter has an interpretation
    as being equal to the *median* lifetime. The
    :math:`\beta` parameter influences the shape of the hazard.

    Parameters
    ----------
    alpha_cols: pd.Index or coercible, optional, default=None
        Columns of the input data frame to be used as covariates for
        the scale parameter :math:`\alpha`.
        If None, all columns are used.

    beta_cols: string "all", pd.Index or coercible, optional, default=None
        Columns of the input data frame to be used as covariates for
        the shape parameter :math:`\beta`.
        If None, no covariates are used, the shape parameter
        is estimated as a constant. If "all", all columns are used.

    fit_intercept: boolean, optional (default=True)
        Whether to fit an intercept term in the model.

    alpha: float, optional (default=0.05)
      the level in the confidence intervals around the estimated survival function,
      for computation of ``confidence_intervals_`` fitted parameter.

    penalizer: float or array, optional (default=0.0)
        the penalizer coefficient to the size of the coefficients.
        See ``l1_ratio``. Must be equal to or greater than 0.
        Alternatively, penalizer is an array equal in size to the number of parameters,
        with penalty coefficients for specific variables. For
        example, ``penalizer=0.01 * np.ones(p)`` is the same as ``penalizer=0.01``

    l1_ratio: float, optional (default=0.0)
        how much of the penalizer should be attributed to an l1 penalty
        (otherwise an l2 penalty). The penalty function looks like
        ``penalizer * l1_ratio * ||w||_1 + 0.5 * penalizer * (1 - l1_ratio) * ||w||^2_2``  # noqa E501

    Attributes
    ----------
    params_ : DataFrame
        The estimated coefficients
    confidence_intervals_ : DataFrame
        The lower and upper confidence intervals for the coefficients
    durations: Series
        The event_observed variable provided
    event_observed: Series
        The event_observed variable provided
    weights: Series
        The event_observed variable provided
    variance_matrix_ : DataFrame
        The variance matrix of the coefficients
    standard_errors_: Series
        the standard errors of the estimates
    score_: float
        the concordance index of the model.
    """

    _tags = {"authors": ["CamDavidsonPilon", "fkiraly"]}
    # CamDavidsonPilon, credit for interfaced estimator

    def __init__(
        self,
        alpha_cols=None,
        beta_cols=None,
        fit_intercept: bool = True,
        alpha: float = 0.05,
        penalizer: float = 0.0,
        l1_ratio: float = 0.0,
    ):
        self.alpha_cols = alpha_cols
        self.beta_cols = beta_cols
        self.alpha = alpha
        self.penalizer = penalizer
        self.l1_ratio = l1_ratio
        self.fit_intercept = fit_intercept

        super().__init__()

        if alpha_cols is not None:
            self.X_col_subset = alpha_cols

    def _get_lifelines_class(self):
        """Getter of the lifelines class to be used for the adapter."""
        from lifelines.fitters.log_logistic_aft_fitter import LogLogisticAFTFitter

        return LogLogisticAFTFitter

    def _get_lifelines_object(self):
        """Abstract method to initialize lifelines object.

        The default initializes result of _get_lifelines_class
        with self.get_params.
        """
        cls = self._get_lifelines_class()
        params = self.get_params()
        params.pop("alpha_cols", None)
        params.pop("beta_cols", None)
        return cls(**params)

    def _add_extra_fit_args(self, X, y, C=None):
        """Get extra arguments for the fit method.

        Parameters
        ----------
        X : pd.DataFrame
            Training features
        y: pd.DataFrame
            Training labels
        C: pd.DataFrame, optional (default=None)
            Censoring information for survival analysis.
        fit_args: dict, optional (default=None)
            Existing arguments for the fit method, from the adapter.

        Returns
        -------
        dict
            Extra arguments for the fit method.
        """
        if self.alpha_cols is not None:
            if self.alpha_cols == "all":
                return {"ancillary": True}
            else:
                return {"ancillary": X[self.alpha_cols]}
        else:
            return {}

    def _predict_proba(self, X):
        """Predict_proba method adapter.

        Parameters
        ----------
        X : pd.DataFrame
            Features to predict on.

        Returns
        -------
        skpro Empirical distribution
        """
        if self.beta_cols == "all":
            ancillary = X
        elif self.beta_cols is not None:
            ancillary = X[self.beta_cols]
        else:
            ancillary = None

        if self.alpha_cols is not None:
            df = X[self.alpha_cols]
        else:
            df = X

        lifelines_est = getattr(self, self._estimator_attr)
        ll_pred_proba = lifelines_est._prep_inputs_for_prediction_and_return_scores

        alpha, beta = ll_pred_proba(df, ancillary)
        alpha = np.expand_dims(alpha, axis=1)
        beta = np.expand_dims(beta, axis=1)

        dist = Fisk(alpha=alpha, beta=beta, index=X.index, columns=self._y_cols)
        return dist

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return `"default"` set.

        Returns
        -------
        params : dict or list of dict, default = {}
            Parameters to create testing instances of the class
            Each dict are parameters to construct an "interesting" test instance, i.e.,
            `MyClass(**params)` or `MyClass(**params[i])` creates a valid test instance.
            `create_test_instance` uses the first (or only) dictionary in `params`
        """
        params1 = {}

        params2 = {
            "beta_cols": "all",
            "fit_intercept": False,
            "alpha": 0.1,
            "penalizer": 0.1,
            "l1_ratio": 0.1,
        }
        return [params1, params2]



"""Automated tests based on the skbase test suite template."""
import numbers
import types
from copy import deepcopy
from inspect import getfullargspec, isclass, signature

import joblib
import numpy as np
import pandas as pd
from skbase.testing import BaseFixtureGenerator as _BaseFixtureGenerator
from skbase.testing import QuickTester as _QuickTester
from skbase.testing import TestAllObjects as _TestAllObjects
from skbase.testing.utils.inspect import _get_args

from skpro.registry import OBJECT_TAG_LIST, all_objects
from skpro.tests._config import EXCLUDE_ESTIMATORS, EXCLUDED_TESTS
from skpro.tests.scenarios.scenarios_getter import retrieve_scenarios
from skpro.tests.test_switch import run_test_for_class
from skpro.utils._doctest import run_doctest
from skpro.utils.deep_equals import deep_equals
from skpro.utils.random_state import set_random_state

# whether to test only estimators from modules that are changed w.r.t. main
# default is False, can be set to True by pytest --only_changed_modules True flag
ONLY_CHANGED_MODULES = False


class PackageConfig:
    """Contains package config variables for test classes."""

    # class variables which can be overridden by descendants
    # ------------------------------------------------------

    # package to search for objects
    # expected type: str, package/module name, relative to python environment root
    package_name = "skpro"

    # list of object types (class names) to exclude
    # expected type: list of str, str are class names
    exclude_objects = EXCLUDE_ESTIMATORS

    # list of tests to exclude
    # expected type: dict of lists, key:str, value: List[str]
    # keys are class names of estimators, values are lists of test names to exclude
    excluded_tests = EXCLUDED_TESTS

    # list of valid tags
    # expected type: list of str, str are tag names
    valid_tags = OBJECT_TAG_LIST


class BaseFixtureGenerator(_BaseFixtureGenerator):
    """Fixture generator for base testing functionality in sktime.

    Test classes inheriting from this and not overriding pytest_generate_tests
        will have estimator and scenario fixtures parametrized out of the box.

    Descendants can override:
        estimator_type_filter: str, class variable; None or scitype string
            e.g., "forecaster", "transformer", "classifier", see BASE_CLASS_SCITYPE_LIST
            which estimators are being retrieved and tested
        fixture_sequence: list of str
            sequence of fixture variable names in conditional fixture generation
        _generate_[variable]: object methods, all (test_name: str, **kwargs) -> list
            generating list of fixtures for fixture variable with name [variable]
                to be used in test with name test_name
            can optionally use values for fixtures earlier in fixture_sequence,
                these must be input as kwargs in a call
        is_excluded: static method (test_name: str, est: class) -> bool
            whether test with name test_name should be excluded for estimator est
                should be used only for encoding general rules, not individual skips
                individual skips should go on the EXCLUDED_TESTS list in _config
            requires _generate_object_class and _generate_object_instance as is
        _excluded_scenario: static method (test_name: str, scenario) -> bool
            whether scenario should be skipped in test with test_name test_name
            requires _generate_estimator_scenario as is

    Fixtures parametrized
    ---------------------
    object_class: estimator inheriting from BaseObject
        ranges over estimator classes not excluded by EXCLUDE_ESTIMATORS, EXCLUDED_TESTS
    object_instance: instance of estimator inheriting from BaseObject
        ranges over estimator classes not excluded by EXCLUDE_ESTIMATORS, EXCLUDED_TESTS
        instances are generated by create_test_instance class method of object_class
    scenario: instance of TestScenario
        ranges over all scenarios returned by retrieve_scenarios
        applicable for object_class or object_instance
    """

    # overrides object retrieval in scikit-base
    def _all_objects(self):
        """Retrieve list of all object classes of type self.object_type_filter.

        If self.object_type_filter is None, retrieve all objects.
        If class, retrieve all classes inheriting from self.object_type_filter.
        Otherwise (assumed str or list of str), retrieve all classes with tags
        object_type in self.object_type_filter.
        """
        filter = getattr(self, "object_type_filter", None)

        if isclass(filter):
            object_types = filter.get_class_tag("object_type", None)
        else:
            object_types = filter

        obj_list = all_objects(
            object_types=object_types,
            return_names=False,
            exclude_objects=self.exclude_objects,
        )

        if isclass(filter):
            obj_list = [obj for obj in obj_list if issubclass(obj, filter)]

        # run_test_for_class selects the estimators to run
        # based on whether they have changed, and whether they have all dependencies
        # internally, uses the ONLY_CHANGED_MODULES flag,
        # and checks the python env against python_dependencies tag
        obj_list = [obj for obj in obj_list if run_test_for_class(obj)]

        return obj_list

    # which sequence the conditional fixtures are generated in
    fixture_sequence = [
        "object_class",
        "object_instance",
        "scenario",
    ]

    def _generate_scenario(self, test_name, **kwargs):
        """Return estimator test scenario.

        Fixtures parametrized
        ---------------------
        scenario: instance of TestScenario
            ranges over all scenarios returned by retrieve_scenarios
        """
        if "object_class" in kwargs.keys():
            obj = kwargs["object_class"]
        elif "object_instance" in kwargs.keys():
            obj = kwargs["object_instance"]
        else:
            return []

        scenarios = retrieve_scenarios(obj)
        scenarios = [s for s in scenarios if not self._excluded_scenario(test_name, s)]
        scenario_names = [type(scen).__name__ for scen in scenarios]

        return scenarios, scenario_names

    @staticmethod
    def _excluded_scenario(test_name, scenario):
        """Skip list generator for scenarios to skip in test_name.

        Arguments
        ---------
        test_name : str, name of test
        scenario : instance of TestScenario, to be used in test

        Returns
        -------
        bool, whether scenario should be skipped in test_name
        """
        # this line excludes all scenarios that do not have "is_enabled" flag
        #   we should slowly enable more scenarios for better coverage
        # comment out to run the full test suite with new scenarios
        if not scenario.get_tag("is_enabled", False, raise_error=False):
            return True

        return False


class TestAllObjects(PackageConfig, BaseFixtureGenerator, _TestAllObjects):
    """Generic tests for all objects in the mini package."""

    def test_doctest_examples(self, object_class):
        """Runs doctests for estimator class."""
        run_doctest(object_class, name=f"class {object_class.__name__}")

    # override this due to reserved_params index, columns, in the BaseDistribution class
    # index and columns params behave like pandas, i.e., are changed after __init__
    def test_constructor(self, object_class):
        """Check that the constructor has sklearn compatible signature and behaviour.

        Based on sklearn check_estimator testing of __init__ logic.
        Uses create_test_instance to create an instance.
        Assumes test_create_test_instance has passed and certified create_test_instance.

        Tests that:
        * constructor has no varargs
        * tests that constructor constructs an instance of the class
        * tests that all parameters are set in init to an attribute of the same name
        * tests that parameter values are always copied to the attribute and not changed
        * tests that default parameters are one of the following:
            None, str, int, float, bool, tuple, function, joblib memory, numpy primitive
            (other type parameters should be None, default handling should be by writing
            the default to attribute of a different name, e.g., my_param_ not my_param)
        """
        msg = "constructor __init__ should have no varargs"
        assert getfullargspec(object_class.__init__).varkw is None, msg

        estimator = object_class.create_test_instance()
        assert isinstance(estimator, object_class)

        # Ensure that each parameter is set in init
        init_params = _get_args(type(estimator).__init__)
        invalid_attr = set(init_params) - set(vars(estimator)) - {"self"}
        assert not invalid_attr, (
            "Estimator %s should store all parameters"
            " as an attribute during init. Did not find "
            "attributes `%s`." % (estimator.__class__.__name__, sorted(invalid_attr))
        )

        # Ensure that init does nothing but set parameters
        # No logic/interaction with other parameters
        def param_filter(p):
            """Identify hyper parameters of an estimator."""
            return p.name != "self" and p.kind not in [p.VAR_KEYWORD, p.VAR_POSITIONAL]

        init_params = [
            p
            for p in signature(estimator.__init__).parameters.values()
            if param_filter(p)
        ]

        params = estimator.get_params()

        test_params = object_class.get_test_params()
        if isinstance(test_params, list):
            test_params = test_params[0]
        test_params = test_params.keys()

        init_params = [param for param in init_params if param.name not in test_params]

        for param in init_params:
            assert param.default != param.empty, (
                "parameter `%s` for %s has no default value and is not "
                "set in `get_test_params`" % (param.name, estimator.__class__.__name__)
            )
            if type(param.default) is type:
                assert param.default in [np.float64, np.int64]
            else:
                assert type(param.default) in [
                    str,
                    int,
                    float,
                    bool,
                    tuple,
                    type(None),
                    np.float64,
                    types.FunctionType,
                ]

            reserved_params = object_class.get_class_tag("reserved_params", [])
            if param.name not in reserved_params:
                param_value = params[param.name]
                if isinstance(param_value, np.ndarray):
                    np.testing.assert_array_equal(param_value, param.default)
                elif bool(
                    isinstance(param_value, numbers.Real) and np.isnan(param_value)
                ):
                    # Allows to set default parameters to np.nan
                    assert param_value is param.default, param.name
                else:
                    assert param_value == param.default, param.name

    # same here, reserved_params need to be dealt with
    def test_set_params_sklearn(self, object_class):
        """Check that set_params works correctly, mirrors sklearn check_set_params.

        Instead of the "fuzz values" in sklearn's check_set_params,
        we use the other test parameter settings (which are assumed valid).
        This guarantees settings which play along with the __init__ content.
        """
        from skpro.utils.deep_equals import deep_equals

        estimator = object_class.create_test_instance()
        test_params = object_class.get_test_params()
        if not isinstance(test_params, list):
            test_params = [test_params]

        reserved_params = object_class.get_class_tag(
            "reserved_params", tag_value_default=[]
        )

        for params in test_params:
            # we construct the full parameter set for params
            # params may only have parameters that are deviating from defaults
            # in order to set non-default parameters back to defaults
            params_full = object_class.get_param_defaults()
            params_full.update(params)

            msg = f"set_params of {object_class.__name__} does not return self"
            est_after_set = estimator.set_params(**params_full)
            assert est_after_set is estimator, msg

            def unreserved(params):
                return {p: v for p, v in params.items() if p not in reserved_params}

            est_params = estimator.get_params(deep=False)
            is_equal, equals_msg = deep_equals(
                unreserved(est_params), unreserved(params_full), return_msg=True
            )
            msg = (
                f"get_params result of {object_class.__name__} (x) does not match "
                f"what was passed to set_params (y). "
                f"Reason for discrepancy: {equals_msg}"
            )
            assert is_equal, msg

    def test_get_test_params_coverage(self, object_class):
        """Check that get_test_params has good test coverage.

        Checks that:

        * get_test_params returns at least two test parameter sets
        """
        param_list = object_class.get_test_params()

        if isinstance(param_list, dict):
            param_list = [param_list]

        def _coerce_to_list_of_str(obj):
            if isinstance(obj, str):
                return obj
            elif isinstance(obj, list):
                return obj
            else:
                return []

        reserved_param_names = object_class.get_class_tag(
            "reserved_params", tag_value_default=None
        )
        reserved_param_names = _coerce_to_list_of_str(reserved_param_names)
        reserved_set = set(reserved_param_names)

        param_names = object_class.get_param_names()
        unreserved_param_names = set(param_names).difference(reserved_set)

        if len(unreserved_param_names) > 0:
            msg = (
                f"{object_class.__name__}.get_test_params should return "
                f"at least two test parameter sets, but only {len(param_list)} found."
            )
            assert len(param_list) > 1, msg


class TestAllEstimators(PackageConfig, BaseFixtureGenerator, _QuickTester):
    """Package level tests for all sktime estimators, i.e., objects with fit."""

    def test_fit_updates_state(self, object_instance, scenario):
        """Check fit/update state change."""
        # Check that fit updates the is-fitted states
        attrs = ["_is_fitted", "is_fitted"]

        estimator = object_instance
        object_class = type(object_instance)

        msg = (
            f"{object_class.__name__}.__init__ should call "
            f"super({object_class.__name__}, self).__init__, "
            "but that does not seem to be the case. Please ensure to call the "
            f"parent class's constructor in {object_class.__name__}.__init__"
        )
        assert hasattr(estimator, "_is_fitted"), msg

        # Check is_fitted attribute is set correctly to False before fit, at init
        for attr in attrs:
            assert not getattr(
                estimator, attr
            ), f"Estimator: {estimator} does not initiate attribute: {attr} to False"

        fitted_estimator = scenario.run(object_instance, method_sequence=["fit"])

        # Check is_fitted attributes are updated correctly to True after calling fit
        for attr in attrs:
            assert getattr(
                fitted_estimator, attr
            ), f"Estimator: {estimator} does not update attribute: {attr} during fit"

    def test_fit_returns_self(self, object_instance, scenario):
        """Check that fit returns self."""
        fit_return = scenario.run(object_instance, method_sequence=["fit"])
        assert (
            fit_return is object_instance
        ), f"Estimator: {object_instance} does not return self when calling fit"

    def test_fit_does_not_overwrite_hyper_params(self, object_instance, scenario):
        """Check that we do not overwrite hyper-parameters in fit."""
        estimator = object_instance
        set_random_state(estimator)

        # Make a physical copy of the original estimator parameters before fitting.
        params = estimator.get_params()
        original_params = deepcopy(params)

        # Fit the model
        fitted_est = scenario.run(object_instance, method_sequence=["fit"])

        # Compare the state of the model parameters with the original parameters
        new_params = fitted_est.get_params()
        for param_name, original_value in original_params.items():
            new_value = new_params[param_name]

            # We should never change or mutate the internal state of input
            # parameters by default. To check this we use the joblib.hash function
            # that introspects recursively any subobjects to compute a checksum.
            # The only exception to this rule of immutable constructor parameters
            # is possible RandomState instance but in this check we explicitly
            # fixed the random_state params recursively to be integer seeds.
            msg = (
                "Estimator %s should not change or mutate "
                " the parameter %s from %s to %s during fit."
                % (estimator.__class__.__name__, param_name, original_value, new_value)
            )
            # joblib.hash has problems with pandas objects, so we use deep_equals then
            if isinstance(original_value, (pd.DataFrame, pd.Series)):
                assert deep_equals(new_value, original_value), msg
            else:
                assert joblib.hash(new_value) == joblib.hash(original_value), msg



# copyright: sktime developers, BSD-3-Clause License (see LICENSE file)
"""Uniform probability distribution."""

__author__ = ["an20805"]

import numpy as np
import pandas as pd

from skpro.distributions.base import BaseDistribution


class Uniform(BaseDistribution):
    r"""Continuous uniform distribution.

    The uniform distribution is parameterized by lower and upper bounds of interval,
    :math:`a` and :math`b`, such that the pdf is

    .. math:: f(x) = \frac{1}{b - a} \text{ for } a \leq x \leq b, \text{ and } 0 \text{ otherwise}  # noqa E501

    The lower bound :math:`a` is represented by the parameter ``lower``,
    and the upper bound :math:`b` by the parameter ``upper``.

    Parameters
    ----------
    lower : float
        Lower bound of the distribution.
    upper : float, must be greater than lower
        Upper bound of the distribution.
    index : pd.Index, optional, default = RangeIndex
    columns : pd.Index, optional, default = RangeIndex

    Examples
    --------
    >>> from skpro.distributions import Uniform

    >>> u = Uniform(lower=0, upper=5)
    """

    _tags = {
        "authors": ["an20805"],
        "capabilities:approx": ["pdfnorm"],
        "capabilities:exact": ["pdf", "log_pdf", "cdf", "ppf", "mean", "var", "energy"],
        "distr:measuretype": "continuous",
        "distr:paramtype": "parametric",
        "broadcast_init": "on",
    }

    def __init__(self, lower, upper, index=None, columns=None):
        self.lower = lower
        self.upper = upper

        super().__init__(index=index, columns=columns)

        if self.ndim == 0 and lower >= upper:
            raise ValueError(
                "Error in Uniform distribution parameters, "
                "upper bound must be strictly greater than "
                "lower bound."
            )
        else:
            # use 2D broadcasted params for checking
            lower = self._bc_params["lower"]
            upper = self._bc_params["upper"]

            if np.any(lower >= upper):
                raise ValueError(
                    "Error in Uniform distribution parameters, "
                    "upper bound must be strictly greater than "
                    "lower bound."
                )

    def _pdf(self, x):
        """Probability density function.

        Parameters
        ----------
        x : 2D np.ndarray, same shape as ``self``
            values to evaluate the pdf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            pdf values at the given points
        """
        lower = self._bc_params["lower"]
        upper = self._bc_params["upper"]

        in_bounds = np.logical_and(x >= lower, x <= upper)
        pdf_arr = in_bounds / (upper - lower)
        return pdf_arr

    def _cdf(self, x):
        """Cumulative distribution function.

        Parameters
        ----------
        x : 2D np.ndarray, same shape as ``self``
            values to evaluate the cdf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            cdf values at the given points
        """
        lower = self._bc_params["lower"]
        upper = self._bc_params["upper"]

        in_bounds = (x >= lower) & (x <= upper)
        above_bound = x > upper

        cdf_arr = in_bounds * (x - lower) / (upper - lower) + above_bound
        return cdf_arr

    def _ppf(self, p):
        """Quantile function = percent point function = inverse cdf.

        Parameters
        ----------
        p : 2D np.ndarray, same shape as ``self``
            values to evaluate the ppf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            ppf values at the given points
        """
        lower = self._bc_params["lower"]
        upper = self._bc_params["upper"]

        ppf_arr = lower + p * (upper - lower)
        return ppf_arr

    def _mean(self):
        """Return expected value of the distribution.

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            expected value of distribution (entry-wise)
        """
        lower = self._bc_params["lower"]
        upper = self._bc_params["upper"]

        mean_arr = (lower + upper) / 2
        return mean_arr

    def _var(self):
        r"""Return element/entry-wise variance of the distribution.

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            variance of the distribution (entry-wise)
        """
        lower = self._bc_params["lower"]
        upper = self._bc_params["upper"]

        var_arr = (upper - lower) ** 2 / 12
        return var_arr

    def _energy_self(self):
        r"""Energy of self, w.r.t. self.

        :math:`\mathbb{E}[|X-Y|]`, where :math:`X, Y` are i.i.d. copies of self.

        Private method, to be implemented by subclasses.

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            energy values w.r.t. the given points
        """
        lower = self._bc_params["lower"]
        upper = self._bc_params["upper"]

        energy_arr = (upper - lower) / 3  # Expected absolute difference

        if energy_arr.ndim > 0:
            energy_arr = np.sum(energy_arr, axis=1)
        return energy_arr

    def _energy_x(self, x):
        r"""Energy of self, w.r.t. a constant frame x.

        :math:`\mathbb{E}[|X-x|]`, where :math:`X` is a copy of self,
        and :math:`x` is a constant.

        Private method, to be implemented by subclasses.

        Parameters
        ----------
        x : 2D np.ndarray, same shape as ``self``
            values to compute energy w.r.t. to

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            energy values w.r.t. the given points
        """
        a = self._bc_params["lower"]
        b = self._bc_params["upper"]

        is_outside = np.logical_or(x < a, x > b)
        is_inside = 1 - is_outside

        midpoint = (a + b) / 2
        energy_arr = is_outside * np.abs(x - midpoint)
        energy_arr += is_inside * ((b - x) ** 2 + (a - x) ** 2) / (2 * (b - a))

        if energy_arr.ndim > 0:
            energy_arr = np.sum(energy_arr, axis=1)
        return energy_arr

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator."""
        # array case examples
        params1 = {"lower": 0, "upper": [5, 10]}
        params2 = {
            "lower": -5,
            "upper": 5,
            "index": pd.Index([1, 3, 5]),
            "columns": pd.Index(["a", "b"]),
        }
        # scalar case examples
        params3 = {"lower": 0, "upper": 3}

        return [params1, params2, params3]



"""Utilities for handling the random_state variable."""
# copied from scikit-learn to avoid dependency on sklearn private methods

import numpy as np
from sklearn.utils import check_random_state


def set_random_state(estimator, random_state=0):
    """Set fixed random_state parameters for an estimator.

    Finds all parameters ending ``random_state`` and sets them to integers
    derived from ``random_state``.

    Parameters
    ----------
    estimator : estimator supporting get_params, set_params
        Estimator with potential randomness managed by random_state parameters.

    random_state : int, RandomState instance or None, default=None
        Pseudo-random number generator to control the generation of the random
        integers. Pass an int for reproducible output across multiple function calls.

    Notes
    -----
    This does not necessarily set *all* ``random_state`` attributes that
    control an estimator's randomness, only those accessible through
    ``estimator.get_params()``.
    """
    random_state = check_random_state(random_state)
    to_set = {}
    for key in sorted(estimator.get_params(deep=True)):
        if key == "random_state" or key.endswith("__random_state"):
            to_set[key] = random_state.randint(np.iinfo(np.int32).max)

    if to_set:
        estimator.set_params(**to_set)



# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
"""Adapter for Scipy Distributions."""

__author__ = ["malikrafsan"]

from typing import Union

import pandas as pd
from scipy.stats import rv_continuous, rv_discrete

from skpro.distributions.base import BaseDistribution

__all__ = ["_ScipyAdapter"]


class _ScipyAdapter(BaseDistribution):
    """Adapter for scipy distributions.

    This class is an adapter for scipy distributions. It provides a common
    interface for all scipy distributions. The class is abstract
    and should not be instantiated directly.
    """

    _distribution_attr = "_dist"
    _tags = {
        "object_type": "distribution",
        "distr:paramtype": "parametric",
    }

    def __init__(self, index=None, columns=None):
        obj = self._get_scipy_object()
        setattr(self, self._distribution_attr, obj)
        super().__init__(index, columns)

    def _get_scipy_object(self) -> Union[rv_continuous, rv_discrete]:
        """Abstract method to get the scipy distribution object.

        Should import the scipy distribution object and return it.
        """
        raise NotImplementedError("abstract method")

    def _get_scipy_param(self):
        """Abstract method to get the scipy distribution parameters.

        Should return a tuple with two elements: a list of positional arguments (args)
        and a dictionary of keyword arguments (kwds).
        """
        raise NotImplementedError("abstract method")

    def _mean(self):
        obj: Union[rv_continuous, rv_discrete] = getattr(self, self._distribution_attr)
        args, kwds = self._get_scipy_param()
        return obj.mean(*args, **kwds)

    def _var(self):
        obj: Union[rv_continuous, rv_discrete] = getattr(self, self._distribution_attr)
        args, kwds = self._get_scipy_param()
        return obj.var(*args, **kwds)

    def _pdf(self, x: pd.DataFrame):
        obj: Union[rv_continuous, rv_discrete] = getattr(self, self._distribution_attr)
        args, kwds = self._get_scipy_param()
        return obj.pdf(x, *args, **kwds)

    def _log_pdf(self, x: pd.DataFrame):
        obj: Union[rv_continuous, rv_discrete] = getattr(self, self._distribution_attr)
        args, kwds = self._get_scipy_param()
        return obj.logpdf(x, *args, **kwds)

    def _cdf(self, x: pd.DataFrame):
        obj: Union[rv_continuous, rv_discrete] = getattr(self, self._distribution_attr)
        args, kwds = self._get_scipy_param()
        return obj.cdf(x, *args, **kwds)

    def _ppf(self, p: pd.DataFrame):
        obj: Union[rv_continuous, rv_discrete] = getattr(self, self._distribution_attr)
        args, kwds = self._get_scipy_param()
        return obj.ppf(p, *args, **kwds)

    def _pmf(self, x: pd.DataFrame):
        """Return the probability mass function evaluated at x."""
        obj: Union[rv_continuous, rv_discrete] = getattr(self, self._distribution_attr)
        args, kwds = self._get_scipy_param()
        return obj.pmf(x, *args, **kwds)

    def _log_pmf(self, x: pd.DataFrame):
        """Return the log of the probability mass function evaluated at x."""
        obj: Union[rv_continuous, rv_discrete] = getattr(self, self._distribution_attr)
        args, kwds = self._get_scipy_param()
        return obj.logpmf(x, *args, **kwds)



"""Delegator mixin that delegates all methods to wrapped regressor.

Useful for building estimators where all but one or a few methods are delegated. For
that purpose, inherit from this estimator and then override only the methods that
are not delegated.
"""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)

__author__ = ["fkiraly"]
__all__ = ["_DelegatedProbaRegressor"]

from skpro.regression.base._base import BaseProbaRegressor


class _DelegatedProbaRegressor(BaseProbaRegressor):
    """Delegator mixin that delegates all methods to wrapped regressor.

    Delegates inner methods to a wrapped estimator.
        Wrapped estimator is value of attribute with name self._delegate_name.
        By default, this is "estimator_", i.e., delegates to self.estimator_
        To override delegation, override _delegate_name attribute in child class.

    Delegates the following inner underscore methods:
        _fit, _predict,
        _predict_interval, _predict_quantiles, _predict_var, _predict_proba

    Does NOT delegate get_params, set_params.
        get_params, set_params will hence use one additional nesting level by default.

    Does NOT delegate or copy tags, this should be done in a child class if required.
    """

    # attribute for _DelegatedProbaRegressor, which then delegates
    #     all non-overridden methods are same as of getattr(self, _delegate_name)
    #     see further details in _DelegatedRegressor docstring
    _delegate_name = "estimator_"

    def _get_delegate(self):
        return getattr(self, self._delegate_name)

    def _fit(self, X, y, C=None):
        """Fit regressor to training data.

        Writes to self:
            Sets fitted model attributes ending in "_".

        Parameters
        ----------
        X : pandas DataFrame
            feature instances to fit regressor to
        y : pd.DataFrame, must be same length as X
            labels to fit regressor to
        C : pd.DataFrame, optional (default=None)
            censoring information for survival analysis,
            should have same column name as y, same length as X and y
            should have entries 0 and 1 (float or int)
            0 = uncensored, 1 = (right) censored
            if None, all observations are assumed to be uncensored
            Can be passed to any probabilistic regressor,
            but is ignored if capability:survival tag is False.

        Returns
        -------
        self : reference to self
        """
        estimator = self._get_delegate()
        estimator.fit(X=X, y=y, C=C)
        return self

    def _update(self, X, y, C=None):
        """Update regressor with a new batch of training data.

        State required:
            Requires state to be "fitted" = self.is_fitted=True

        Writes to self:
            Updates fitted model attributes ending in "_".

        Parameters
        ----------
        X : pandas DataFrame
            feature instances to fit regressor to
        y : pd.DataFrame, must be same length as X
            labels to fit regressor to
        C : pd.DataFrame, optional (default=None)
            censoring information for survival analysis,
            should have same column name as y, same length as X and y
            should have entries 0 and 1 (float or int)
            0 = uncensored, 1 = (right) censored
            if None, all observations are assumed to be uncensored
            Can be passed to any probabilistic regressor,
            but is ignored if capability:survival tag is False.

        Returns
        -------
        self : reference to self
        """
        estimator = self._get_delegate()
        estimator.update(X=X, y=y, C=C)
        return self

    def _predict(self, X):
        """Predict labels for data from features.

        State required:
            Requires state to be "fitted" = self.is_fitted=True

        Accesses in self:
            Fitted model attributes ending in "_"

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for

        Returns
        -------
        y : pandas DataFrame, same length as `X`, same columns as `y` in `fit`
            labels predicted for `X`
        """
        estimator = self._get_delegate()
        return estimator.predict(X=X)

    def _predict_quantiles(self, X, alpha):
        """Compute/return quantile predictions.

        private _predict_quantiles containing the core logic,
            called from predict_quantiles and default _predict_interval

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for
        alpha : guaranteed list of float
            A list of probabilities at which quantile predictions are computed.

        Returns
        -------
        quantiles : pd.DataFrame
            Column has multi-index: first level is variable name from ``y`` in fit,
                second level being the values of alpha passed to the function.
            Row index is equal to row index of ``X``.
            Entries are quantile predictions, for var in col index,
                at quantile probability in second col index, for the row index.
        """
        estimator = self._get_delegate()
        return estimator.predict_quantiles(X=X, alpha=alpha)

    def _predict_interval(self, X, coverage):
        """Compute/return interval predictions.

        private _predict_interval containing the core logic,
            called from predict_interval and default _predict_quantiles

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for
        coverage : guaranteed list of float of unique values
           nominal coverage(s) of predictive interval(s)

        Returns
        -------
        pred_int : pd.DataFrame
            Column has multi-index: first level is variable name from ``y`` in fit,
            second level coverage fractions for which intervals were computed,
            in the same order as in input `coverage`.
            Third level is string "lower" or "upper", for lower/upper interval end.
            Row index is equal to row index of ``X``.
            Entries are lower/upper bounds of interval predictions,
            for var in col index, at nominal coverage in second col index,
            lower/upper depending on third col index, for the row index.
            Upper/lower interval end are equivalent to
            quantile predictions at alpha = 0.5 - c/2, 0.5 + c/2 for c in coverage.
        """
        estimator = self._get_delegate()
        return estimator.predict_interval(X=X, coverage=coverage)

    def _predict_var(self, X):
        """Compute/return variance predictions.

        private _predict_var containing the core logic, called from predict_var

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for

        Returns
        -------
        pred_var : pd.DataFrame
            Column names are exactly those of ``y`` passed in ``fit``.
            Row index is equal to row index of ``X``.
            Entries are variance prediction, for var in col index.
            A variance prediction for given variable and fh index is a predicted
            variance for that variable and index, given observed data.
        """
        estimator = self._get_delegate()
        return estimator.predict_var(X=X)

    def _predict_proba(self, X):
        """Predict distribution over labels for data from features.

        State required:
            Requires state to be "fitted".

        Accesses in self:
            Fitted model attributes ending in "_"

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for

        Returns
        -------
        y : skpro BaseDistribution, same length as `X`
            labels predicted for `X`
        """
        estimator = self._get_delegate()
        return estimator.predict_proba(X=X)

    def _get_fitted_params(self):
        """Get fitted parameters.

        private _get_fitted_params, called from get_fitted_params

        State required:
            Requires state to be "fitted".

        Returns
        -------
        fitted_params : dict with str keys
            fitted parameters, keyed by names of fitted parameter
        """
        estimator = self._get_delegate()
        return estimator.get_fitted_params()



"""Tests for Empirical distributions."""

import pandas as pd
import pytest

from skpro.distributions.empirical import Empirical
from skpro.tests.test_switch import run_test_module_changed


@pytest.mark.skipif(
    not run_test_module_changed("skpro.distributions"),
    reason="run only if skpro.distributions has been changed",
)
def test_empirical_iat_index():
    """Test that the index is correctly set after iat call."""
    spl_idx = pd.MultiIndex.from_product([[0, 1], [0, 1, 2]], names=["sample", "time"])
    spl = pd.DataFrame(
        [[0, 1], [2, 3], [10, 11], [6, 7], [8, 9], [4, 5]],
        index=spl_idx,
        columns=["a", "b"],
    )
    emp = Empirical(spl, columns=["a", "b"])

    emp_iat = emp.iat[0, 0]
    assert emp_iat.shape == ()

    assert not isinstance(emp_iat.spl.index, pd.MultiIndex)
    assert (emp_iat.spl.index == [0, 1]).all()



"""Interface adapter for MAPIE regressor."""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)

__author__ = ["fkiraly"]  # interface only. MAPIE authors in mapie package

import numpy as np
import pandas as pd

from skpro.regression.base import BaseProbaRegressor
from skpro.utils.sklearn import prep_skl_df


class MapieRegressor(BaseProbaRegressor):
    """MAPIE probabilistic regressor, conformity score based prediction intervals.

    Direct interface to ``mapie.regression.regression.MapieRegressor`` from the
    ``mapie`` package.

    Uses jackknife+ to estimate prediction intervals on a per-sample basis.

    Any (non-probabilistic) sklearn regressor can be used as the base regressor,
    ``MapieRegressor`` will add prediction intervals.

    Parameters
    ----------
    estimator: sklearn regressor, default = ``sklearn.linear_model.LinearRegression``
        regressor with scikit-learn compatible API
        (requires ``fit``, ``predict``, ``get_params``, ``set_params``).
        If ``None``, estimator defaults to a ``LinearRegression`` instance.

    method: str, default = "plus"
        Method to choose for prediction interval estimates.

        - ``"naive"``, based on training set conformity scores,
        - ``"base"``, based on validation sets conformity scores,
        - ``"plus"``, based on validation conformity scores and
          testing predictions,
        - ``"minmax"``, based on validation conformity scores and
          testing predictions (min/max among cross-validation clones).

    cv: int, str, or sklearn splitter, default = 5-fold cross-validation
        Cross-validation strategy for computing conformity scores.

        - ``None``, to use the default 5-fold cross-validation
        - integer, to specify the number of folds.
          If equal to ``-1``, equivalent to
          ``sklearn.model_selection.LeaveOneOut()``.
        - CV splitter: any ``sklearn.model_selection.BaseCrossValidator``
          Main variants are:
            - ``sklearn.model_selection.LeaveOneOut`` (jackknife),
            - ``sklearn.model_selection.KFold`` (cross-validation),
            - ``subsample.Subsample`` object (bootstrap).
        - ``"split"``, does not involve cross-validation but a division
          of the data into training and calibration subsets. The splitter
          used is the following: ``sklearn.model_selection.ShuffleSplit``.
          ``method`` parameter is set to ``"base"``.
        - ``"prefit"``, assumes that ``estimator`` has been fitted already,
          and the ``method`` parameter is set to ``"base"``.
          All data provided in the ``fit`` method is then used
          for computing conformity scores only.
          At prediction time, quantiles of these conformity scores are used
          to provide a prediction interval with fixed width.
          The user has to take care manually that data for model fitting and
          conformity scores estimate are disjoint.

        By default ``None``.

    test_size: int, float, or None (default).
        Test size for the ``"split"`` cross-validation strategy.
        Ignored unless ``cv`` is ``"split"``.

        If ``float``, must be between ``0.0`` and ``1.0`` and represents the
        proportion of the dataset to include in the test split.
        If ``int``, represents the absolute number of test samples.
        If ``None``, set to ``0.1``.

    n_jobs: int, optional, default=None
        Number of jobs for parallel processing using ``joblib``
        via the "locky" backend.
        If ``-1`` all CPUs are used.
        If ``1`` is given, no parallel computing code is used at all,
        which is useful for debugging.
        For ``n_jobs`` below ``-1``, ``(n_cpus + 1 - n_jobs)`` are used.
        ``None`` is a marker for `unset` that will be interpreted as
        ``n_jobs=1`` (sequential execution).

    agg_function: str, one of "mean" (default), "median", or None.
        Determines how to aggregate predictions from perturbed models, both at
        training and prediction time.

        If ``None``, it is ignored except if ``cv`` class is ``Subsample``,
        in which case an error is raised.
        If ``"mean"`` or ``"median"``, returns the mean or median of the
        predictions computed from the out-of-folds models.
        Note: if you plan to set the ``ensemble`` argument to ``True`` in the
        ``predict`` method, you have to specify an aggregation function.
        Otherwise an error would be raised.

        The Jackknife+ interval can be interpreted as an interval around the
        median prediction, and is guaranteed to lie inside the interval,
        unlike the single estimator predictions.

        When the cross-validation strategy is ``Subsample`` (i.e. for the
        Jackknife+-after-Bootstrap method), this function is also used to
        aggregate the training set in-sample predictions.

        If ``cv`` is ``"prefit"`` or ``"split"``, ``agg_function`` is ignored.

    verbose: int, optional, default=0
        The verbosity level, used with joblib for multiprocessing.
        The frequency of the messages increases with the verbosity level.
        If it more than ``10``, all iterations are reported.
        Above ``50``, the output is sent to stdout.

    conformity_score: ``ConformityScore``, optional, default = absolute conformity score
        mapie ConformityScore descendant instance, from ``mapie.conformity_scores``.
        It defines the link between the observed values, the predicted ones
        and the conformity scores. For instance, the default ``None`` value
        correspondonds to a conformity score which assumes
        y_obs = y_pred + conformity_score.

        - ``None``, to use the default ``AbsoluteConformityScore`` conformity
          score
        - ConformityScore: any ``ConformityScore`` class

    ensemble: bool, optional, default=False
        Boolean determining whether predictions are ensembled or not.
        Ignored if ``cv`` is ``"prefit"`` or ``"split"``.

        If ``False``, predictions areof the model trained on the full training set.
        If ``True``, predictions from perturbed models are aggregated by
        the aggregation function specified in the ``agg_function``attribute.

    random_state: None (default), ``int``, or ``RandomState`` instance
        Pseudo random number generator state used for random sampling.
        Pass an int for reproducible output across multiple function calls.
        If None, no random state is used.

    Attributes
    ----------
    estimator_: EnsembleRegressor
        fitted mapie ensemble regressor

    conformity_scores_: ArrayLike of shape (n_samples_train,)
        Conformity scores between ``y_train`` and ``y_pred``.

    Examples
    --------
    >>> from skpro.regression.mapie import MapieRegressor  # doctest: +SKIP
    >>> from sklearn.ensemble import RandomForestRegressor  # doctest: +SKIP
    >>> from sklearn.datasets import load_diabetes  # doctest: +SKIP
    >>> from sklearn.model_selection import train_test_split  # doctest: +SKIP
    >>>
    >>> X, y = load_diabetes(return_X_y=True, as_frame=True)  # doctest: +SKIP
    >>> X_train, X_test, y_train, y_test = train_test_split(X, y)  # doctest: +SKIP
    >>>
    >>> reg_tabular = RandomForestRegressor()  # doctest: +SKIP
    >>>
    >>> reg_proba = MapieRegressor(reg_tabular)  # doctest: +SKIP
    >>> reg_proba.fit(X_train, y_train)  # doctest: +SKIP
    MapieRegressor(...)
    >>> y_pred_int = reg_proba.predict_interval(X_test)  # doctest: +SKIP
    >>> y_pred_dist = reg_proba.predict_proba(X_test)  # doctest: +SKIP
    """

    _tags = {
        # packaging info
        # --------------
        "authors": ["fkiraly"],
        "python_dependencies": ["MAPIE"],
        # estimator tags
        # --------------
        "capability:missing": True,
    }

    def __init__(
        self,
        estimator=None,
        method="plus",
        cv=None,
        test_size=None,
        n_jobs=None,
        agg_function="mean",
        verbose=0,
        conformity_score=None,
        random_state=None,
    ):
        self.estimator = estimator
        self.method = method
        self.cv = cv
        self.test_size = test_size
        self.n_jobs = n_jobs
        self.agg_function = agg_function
        self.verbose = verbose
        self.conformity_score = conformity_score
        self.random_state = random_state

        super().__init__()

    def _fit(self, X, y):
        """Fit regressor to training data.

        Writes to self:
            Sets fitted model attributes ending in "_".

        Parameters
        ----------
        X : pandas DataFrame
            feature instances to fit regressor to
        y : pandas DataFrame, must be same length as X
            labels to fit regressor to

        Returns
        -------
        self : reference to self
        """
        # construct mapie regressor
        from mapie.regression import MapieRegressor

        PARAMS_TO_FORWARD = [
            "estimator",
            "method",
            "cv",
            "test_size",
            "n_jobs",
            "agg_function",
            "verbose",
            "conformity_score",
            "random_state",
        ]

        params = self.get_params(deep=False)
        params = {k: v for k, v in params.items() if k in PARAMS_TO_FORWARD}
        mapie_est_ = MapieRegressor(**params)

        # remember y columns for predict
        self._y_cols = y.columns

        # coerce y to numpy array
        if len(y.columns) == 1:
            y = y.to_numpy().flatten()

        # coerce X to pandas DataFrame with string column names
        X = prep_skl_df(X, copy_df=True)

        # fit mapie regressor and save to self
        mapie_est_.fit(X, y)

        self.estimator_mapie_ = mapie_est_

        # forward fitted attributes to self
        FITTED_PARAMS_TO_FORWARD = ["estimator_", "conformity_scores_"]

        for param in FITTED_PARAMS_TO_FORWARD:
            setattr(self, param, getattr(self.estimator_mapie_, param))

        return self

    def _predict(self, X):
        """Predict labels for data from features.

        State required:
            Requires state to be "fitted" = self.is_fitted=True

        Accesses in self:
            Fitted model attributes ending in "_"

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for

        Returns
        -------
        y : pandas DataFrame, same length as `X`, same columns as `y` in `fit`
            labels predicted for `X`
        """
        # coerce X to pandas DataFrame with string column names
        X = prep_skl_df(X, copy_df=True)

        # predict with mapie regressor
        y_pred_mapie = self.estimator_mapie_.predict(X)

        # format output as pandas DataFrame with correct indices
        index = X.index
        columns = self._y_cols
        y_pred = pd.DataFrame(y_pred_mapie, index=index, columns=columns)

        return y_pred

    def _predict_interval(self, X, coverage):
        """Compute/return interval predictions.

        private _predict_interval containing the core logic,
            called from predict_interval and default _predict_quantiles

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for
        coverage : guaranteed list of float of unique values
           nominal coverage(s) of predictive interval(s)

        Returns
        -------
        pred_int : pd.DataFrame
            Column has multi-index: first level is variable name from ``y`` in fit,
            second level coverage fractions for which intervals were computed,
            in the same order as in input `coverage`.
            Third level is string "lower" or "upper", for lower/upper interval end.
            Row index is equal to row index of ``X``.
            Entries are lower/upper bounds of interval predictions,
            for var in col index, at nominal coverage in second col index,
            lower/upper depending on third col index, for the row index.
            Upper/lower interval end are equivalent to
            quantile predictions at alpha = 0.5 - c/2, 0.5 + c/2 for c in coverage.
        """
        cov_arr = np.array(coverage)
        mapie_alpha = 1 - cov_arr

        mapie_pred_int = self.estimator_mapie_.predict(X, alpha=mapie_alpha)[1]

        index = X.index
        columns = pd.MultiIndex.from_product(
            [self._y_cols, coverage, ["lower", "upper"]],
        )

        values = np.reshape(mapie_pred_int, (len(index), -1), order="F")
        pred_int = pd.DataFrame(values, index=index, columns=columns)

        return pred_int

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return `"default"` set.

        Returns
        -------
        params : dict or list of dict, default = {}
            Parameters to create testing instances of the class
            Each dict are parameters to construct an "interesting" test instance, i.e.,
            `MyClass(**params)` or `MyClass(**params[i])` creates a valid test instance.
            `create_test_instance` uses the first (or only) dictionary in `params`
        """
        from sklearn.ensemble import RandomForestRegressor

        params1 = {}

        params2 = {
            "estimator": RandomForestRegressor(),
            "method": "base",
            "cv": 2,
            "test_size": 0.2,
            "agg_function": "median",
            "conformity_score": None,
            "random_state": 42,
        }

        return [params1, params2]



"""Cox proportional hazards models."""

from skpro.survival.ensemble._grad_boost_sksurv import (
    SurvGradBoostCompSkSurv,
    SurvGradBoostSkSurv,
)
from skpro.survival.ensemble._ngboost_surv import NGBoostSurvival
from skpro.survival.ensemble._survforest_sksurv import (
    SurvivalForestSkSurv,
    SurvivalForestXtraSkSurv,
)

__all__ = [
    "SurvGradBoostSkSurv",
    "SurvGradBoostCompSkSurv",
    "SurvivalForestSkSurv",
    "SurvivalForestXtraSkSurv",
    "NGBoostSurvival",
]



"""Type checkers, converters and mtype inference for probabilistic return types."""

from skpro.datatypes._proba._check import check_dict as check_dict_Proba
from skpro.datatypes._proba._convert import convert_dict as convert_dict_Proba
from skpro.datatypes._proba._registry import MTYPE_LIST_PROBA, MTYPE_REGISTER_PROBA

__all__ = [
    "check_dict_Proba",
    "convert_dict_Proba",
    "MTYPE_LIST_PROBA",
    "MTYPE_REGISTER_PROBA",
]



"""Tests for BaseDistribution API points."""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
# adapted from sktime

__author__ = ["fkiraly", "Alex-JG3"]

import numpy as np
import pandas as pd
import pytest
from skbase.testing import QuickTester

from skpro.datatypes import check_is_mtype
from skpro.tests.test_all_estimators import BaseFixtureGenerator, PackageConfig
from skpro.utils.index import random_ss_ix


class DistributionFixtureGenerator(BaseFixtureGenerator):
    """Fixture generator for probability distributions.

    Fixtures parameterized
    ----------------------
    object_class: object inheriting from BaseObject
        ranges over object classes not excluded by EXCLUDE_OBJECTS, EXCLUDED_TESTS
    object_instance: instance of object inheriting from BaseObject
        ranges over object classes not excluded by EXCLUDE_OBJECTS, EXCLUDED_TESTS
        instances are generated by create_test_instance class method
    """

    object_type_filter = "distribution"


def _has_capability(distr, method):
    """Check whether distr has capability of method.

    Parameters
    ----------
    distr : BaseDistribution object
    method : str
        method name to check

    Returns
    -------
    whether distr has capability method, according to tags
    capabilities:approx and capabilities:exact
    """
    approx_methods = distr.get_tag("capabilities:approx")
    exact_methods = distr.get_tag("capabilities:exact")
    return method in approx_methods or method in exact_methods


METHODS_SCALAR = ["mean", "var", "energy"]
METHODS_SCALAR_POS = ["var", "energy"]  # result always non-negative?
METHODS_X = ["energy", "pdf", "log_pdf", "pmf", "log_pmf", "cdf"]
METHODS_X_POS = ["energy", "pdf", "pmf", "cdf", "surv", "haz"]  # result non-negative?
METHODS_P = ["ppf"]
METHODS_ROWWISE = ["energy"]  # results in one column


class TestAllDistributions(PackageConfig, DistributionFixtureGenerator, QuickTester):
    """Module level tests for all skpro parameter fitters."""

    # TEMPORARY skip for CyclicBoosting and QPD classes
    # due to silent failures on main, se #190
    exclude_objects = ["QPD_B"]
    # remove this when fixing failures to re-enable testing

    def test_shape(self, object_instance):
        """Test index, columns, len and shape of distribution."""
        d = object_instance

        assert hasattr(d, "shape")
        assert isinstance(d.shape, tuple)
        assert len(d.shape) in [0, 2]

        if len(d.shape) == 2:
            assert all(isinstance(n, int) for n in d.shape)

            assert isinstance(d.index, pd.Index)
            assert isinstance(d.columns, pd.Index)

            assert d.shape[0] == len(d.index)
            assert d.shape[1] == len(d.columns)

        assert isinstance(len(d), int)

        if len(d.shape) == 2:
            assert len(d) == d.shape[0]
        else:
            assert len(d) == 1

        assert hasattr(d, "ndim")
        assert d.ndim == len(d.shape)

    @pytest.mark.parametrize("shuffled", [False, True])
    def test_sample(self, object_instance, shuffled):
        """Test sample expected return."""
        d = object_instance

        if shuffled:
            d = _shuffle_distr(d)

        res = d.sample()

        if d.ndim > 0:
            assert d.shape == res.shape
            assert (res.index == d.index).all()
            assert (res.columns == d.columns).all()
        else:  # d.ndim = 0
            assert np.isscalar(res)

        res_panel = d.sample(3)
        if d.ndim > 0:
            dummy_panel = pd.concat([res, res, res], keys=range(3))
        else:
            dummy_panel = pd.DataFrame(index=range(3), columns=range(1))
        assert dummy_panel.shape == res_panel.shape
        assert (res_panel.index == dummy_panel.index).all()
        assert (res_panel.columns == dummy_panel.columns).all()

    @pytest.mark.parametrize("shuffled", [False, True])
    @pytest.mark.parametrize("method", METHODS_SCALAR, ids=METHODS_SCALAR)
    def test_methods_scalar(self, object_instance, method, shuffled):
        """Test expected return of scalar methods."""
        if not _has_capability(object_instance, method):
            return None

        d = object_instance
        if shuffled:
            d = _shuffle_distr(d)

        res = getattr(d, method)()

        _check_output_format(res, d, method)

    @pytest.mark.parametrize("shuffled", [False, True])
    @pytest.mark.parametrize("method", METHODS_X, ids=METHODS_X)
    def test_methods_x(self, object_instance, method, shuffled):
        """Test expected return of methods that take sample-like argument."""
        if not _has_capability(object_instance, method):
            return None

        d = object_instance

        if shuffled:
            d = _shuffle_distr(d)

        x = d.sample()
        res = getattr(d, method)(x)

        _check_output_format(res, d, method)

    @pytest.mark.parametrize("shuffled", [False, True])
    @pytest.mark.parametrize("method", METHODS_P, ids=METHODS_P)
    def test_methods_p(self, object_instance, method, shuffled):
        """Test expected return of methods that take percentage-like argument."""
        if not _has_capability(object_instance, method):
            return None

        d = object_instance

        if shuffled:
            d = _shuffle_distr(d)

        np_unif = np.random.uniform(size=d.shape)
        if d.ndim > 0:
            p = pd.DataFrame(np_unif, index=d.index, columns=d.columns)
        else:
            p = np_unif

        res = getattr(d, method)(p)

        _check_output_format(res, d, method)

    @pytest.mark.parametrize("q", [0.7, [0.1, 0.3, 0.9]])
    def test_quantile(self, object_instance, q):
        """Test expected return of quantile method."""
        if not _has_capability(object_instance, "ppf"):
            return None

        d = object_instance

        def _check_quantile_output(obj, q):
            assert check_is_mtype(
                obj, "pred_quantiles", "Proba", msg_return_dict="list"
            )
            if d.ndim == 0:
                expected_index = pd.RangeIndex(1)
                vars = [d.__class__.__name__]
            else:
                expected_index = d.index
                vars = d.columns

            assert (obj.index == expected_index).all()

            if not isinstance(q, list):
                q = [q]
            expected_columns = pd.MultiIndex.from_product([vars, q])
            assert (obj.columns == expected_columns).all()

        res = d.quantile(q)
        _check_quantile_output(res, q)

    @pytest.mark.parametrize("subset_row", [True, False])
    @pytest.mark.parametrize("subset_col", [True, False])
    def test_subsetting(self, object_instance, subset_row, subset_col):
        """Test subsetting of distribution."""
        d = object_instance
        if d.ndim == 0:  # no subsetting to test if example is scalar
            return None

        if subset_row:
            ix_loc = random_ss_ix(d.index, 3)
            ix_iloc = d.index.get_indexer(ix_loc)
        else:
            ix_loc = d.index
            ix_iloc = pd.RangeIndex(len(d.index))

        if subset_col:
            iy_loc = random_ss_ix(d.columns, 1)
            iy_iloc = d.columns.get_indexer(iy_loc)
        else:
            iy_loc = d.columns
            iy_iloc = pd.RangeIndex(len(d.columns))

        res_loc = d.loc[ix_loc, iy_loc]

        assert isinstance(res_loc, type(d))
        assert res_loc.shape == (len(ix_loc), len(iy_loc))
        assert (res_loc.index == ix_loc).all()
        assert (res_loc.columns == iy_loc).all()

        res_iloc = d.iloc[ix_iloc, iy_iloc]

        assert isinstance(res_iloc, type(d))
        assert res_iloc.shape == (len(ix_iloc), len(iy_iloc))
        assert (res_iloc.index == ix_loc).all()
        assert (res_iloc.columns == iy_loc).all()

    def test_log_pdf_and_pdf(self, object_instance):
        """Test that the log of the pdf and log_pdf function are similar."""
        d = object_instance
        capabilities_exact = d.get_tags()["capabilities:exact"]

        if "log_pdf" not in capabilities_exact or "pdf" not in capabilities_exact:
            return
        x = d.sample()
        pdf = d.pdf(x)
        log_pdf = d.log_pdf(x)
        assert np.allclose(np.log(pdf), log_pdf)

    def test_log_pmf_and_pmf(self, object_instance):
        """Test that the log of the pmf and log_pmf function are similar."""
        d = object_instance
        capabilities_exact = d.get_tags()["capabilities:exact"]

        if "log_pmf" not in capabilities_exact or "pmf" not in capabilities_exact:
            return
        x = d.sample()
        pmf = d.pmf(x)
        log_pmf = d.log_pmf(x)
        assert np.allclose(np.log(pmf), log_pmf)

    def test_ppf_and_cdf(self, object_instance):
        """Test that the ppf is the inverse of the cdf."""
        d = object_instance
        capabilities_exact = d.get_tags()["capabilities:exact"]

        if "ppf" not in capabilities_exact or "cdf" not in capabilities_exact:
            return
        x = d.sample()
        x_approx = d.ppf(d.cdf(x))
        if d.ndim > 0:
            assert np.allclose(x.values, x_approx.values)
        else:
            assert np.allclose(x, x_approx)


def _check_output_format(res, dist, method):
    """Check output format expectations for BaseDistribution tests."""
    if dist.shape == ():  # scalar distribution case
        # check if numpy float
        assert np.isscalar(res)
        assert np.isreal(res)
        if method in METHODS_SCALAR_POS or method in METHODS_X_POS:
            assert res >= 0
        return None

    # array distribution case
    if method in METHODS_ROWWISE:
        exp_shape = (dist.shape[0], 1)
    else:
        exp_shape = dist.shape
    assert res.shape == exp_shape
    assert (res.index == dist.index).all()
    if method not in METHODS_ROWWISE:
        assert (res.columns == dist.columns).all()

    if method in METHODS_SCALAR_POS or method in METHODS_X_POS:
        assert (res >= 0).all().all()

    if isinstance(res, pd.DataFrame):
        assert res.apply(pd.api.types.is_numeric_dtype).all()
    elif isinstance(res, pd.Series):
        assert pd.api.types.is_numeric_dtype(res)
    else:
        raise TypeError("res must be a pandas DataFrame or Series.")


def _shuffle_distr(d):
    """Shuffle distribution row index."""
    if d.shape == ():  # nothing to shuffle if scalar
        return d
    # shuffle rows otherwise
    shuffled_df = pd.DataFrame(d.index).sample(frac=1)
    shuffled_index = pd.Index(shuffled_df.values.flatten())
    return d.loc[shuffled_index]



"""Conercion utilities for mtypes."""

__author__ = ["fkiraly"]

import pandas as pd


def _is_nullable_numeric(dtype):
    return dtype in ["Int64", "Float64", "boolean"]


def _coerce_df_dtypes(obj):
    """Coerce pandas objects to non-nullable column types.

    Returns shallow copy and does not mutate input `obj`.

    Parameters
    ----------
    obj: pandas Series or DataFrame, or any object

    Returns
    -------
    obj unchanged, if obj is not pandas Series or DataFrame
    if obj is pandas Series or DataFrame,
        coerces nullable numeric columns to float (by reference via astype)
    """
    if isinstance(obj, pd.Series):
        if _is_nullable_numeric(obj.dtype):
            return obj.astype("float")
        return obj

    if isinstance(obj, pd.DataFrame):
        nullable_cols = [
            col for col in obj.columns if _is_nullable_numeric(obj.dtypes[col])
        ]
        if len(nullable_cols) > 0:
            obj = obj.astype({col: "float" for col in nullable_cols})
        return obj

    return obj



"""Conversion utilities for mtypes."""

__author__ = ["fkiraly"]


def _concat(fun1, fun2):
    """Concatenation of two converter functions, using the same store.

    Parameters
    ----------
    fun1, fun2 : functions in converter signature, see datatypes._convert

    Returns
    -------
    function in converter signature, see datatypes._convert
        concatenation fun2 o fun1, using the same store
    """

    def concat_fun(obj, store=None):
        obj1 = fun1(obj, store=store)
        obj2 = fun2(obj1, store=store)
        return obj2

    return concat_fun


def _extend_conversions(mtype, anchor_mtype, convert_dict, mtype_universe=None):
    """Obtain all conversions from and to mtype via conversion to anchor_mtype.

    Mutates convert_dict by adding all conversions from and to mtype.

    Assumes:
    convert_dict contains
    * conversion from `mtype` to `anchor_mtype`
    * conversion from `anchor_mtype` to `mtype`
    * conversions from `anchor_mtype` to all mtypes in `mtype_universe`
    * conversions from all mtypes in `mtype_universe` to `anchor_mtype`

    Guarantees:
    convert_dict contains
    * conversions from `mtype` to all mtypes in mtype_universe
    * conversions from all mtypes in mtype_universe to `mtype`

    conversions <mtype to mtype2> not in convert_dict at start are filled in as
    _concat(<from mtype to anchor_type>, <from anchor_type to mtype2>)
    conversions <mtype2 to mtype> not in convert_dict at start are filled in as
    _concat(<from mtype2 to anchor_type>, <from anchor_type to mtype>)

    Parameters
    ----------
    mtype : mtype string in convert_dict
    anchor_mtype : mtype string in convert_dict
    convert_dict : conversion dictionary with entries of converter signature
        see docstring of datatypes._convert
    mtype_universe : iterable of mtype strings in convert_dict, coercible to list or set

    Returns
    -------
    reference to convert_dict
    CAVEAT: convert_dict passed to this function gets mutated, this is a reference
    """
    keys = convert_dict.keys()
    scitype = list(keys)[0][2]

    if mtype_universe is None:
        mtype_universe = {x[1] for x in list(keys)}
        mtype_universe = mtype_universe.union([x[0] for x in list(keys)])

    for tp in set(mtype_universe).difference([mtype, anchor_mtype]):
        if (anchor_mtype, tp, scitype) in convert_dict.keys():
            if (mtype, tp, scitype) not in convert_dict.keys():
                convert_dict[(mtype, tp, scitype)] = _concat(
                    convert_dict[(mtype, anchor_mtype, scitype)],
                    convert_dict[(anchor_mtype, tp, scitype)],
                )
        if (tp, anchor_mtype, scitype) in convert_dict.keys():
            if (tp, mtype, scitype) not in convert_dict.keys():
                convert_dict[(tp, mtype, scitype)] = _concat(
                    convert_dict[(tp, anchor_mtype, scitype)],
                    convert_dict[(anchor_mtype, mtype, scitype)],
                )

    return convert_dict



# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
"""Pareto probability distribution."""

__author__ = ["sukjingitsit"]

import numpy as np
import pandas as pd

from skpro.distributions.base import BaseDistribution


class Pareto(BaseDistribution):
    r"""Pareto distribution (skpro native).

    The scale is represented by the parameter ``scale``,
    and the Pareto index (or shape parameter) :math:`\alpha`
    by the parameter ``alpha``.

    The CDF can be represented as

    :math:`F(x) = 1-\left(\frac{\text{scale}}{x}\right)^\alpha
    \text{ if } x>0, 0 \text{ if } x<0`

    Parameters
    ----------
    scale : float or array of float (1D or 2D), must be positive
        scale of the Pareto distribution
    alpha : float or array of float (1D or 2D), must be positive
        shape of the Pareto distribution
    index : pd.Index, optional, default = RangeIndex
    columns : pd.Index, optional, default = RangeIndex

    Examples
    --------
    >>> from skpro.distributions.pareto import Pareto

    >>> n = Pareto(scale=[[1, 1.5], [2, 2.5], [3, 4]], alpha=3)
    """

    _tags = {
        "capabilities:approx": ["pdfnorm", "energy"],
        "capabilities:exact": ["mean", "var", "pdf", "log_pdf", "cdf", "ppf"],
        "distr:measuretype": "continuous",
        "distr:paramtype": "parametric",
        "broadcast_init": "on",
    }

    def __init__(self, scale, alpha, index=None, columns=None):
        self.scale = scale
        self.alpha = alpha

        super().__init__(index=index, columns=columns)

    def _mean(self):
        """Return expected value of the distribution.

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            expected value of distribution (entry-wise)
        """
        alpha = self._bc_params["alpha"]
        scale = self._bc_params["scale"]
        mean = np.where(alpha <= 1, np.inf, scale**alpha / (alpha - 1))
        return mean

    def _var(self):
        r"""Return element/entry-wise variance of the distribution.

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            variance of the distribution (entry-wise)
        """
        alpha = self._bc_params["alpha"]
        scale = self._bc_params["scale"]
        var = np.where(
            alpha <= 2, np.inf, scale**2 * alpha / ((alpha - 2) * (alpha - 1) ** 2)
        )
        return var

    def _pdf(self, x):
        """Probability density function.

        Parameters
        ----------
        x : 2D np.ndarray, same shape as ``self``
            values to evaluate the pdf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            pdf values at the given points
        """
        alpha = self._bc_params["alpha"]
        scale = self._bc_params["scale"]
        pdf_arr = alpha * np.power(scale, alpha)
        pdf_arr /= np.power(x, alpha + 1)
        return pdf_arr

    def _log_pdf(self, x):
        """Logarithmic probability density function.

        Parameters
        ----------
        x : 2D np.ndarray, same shape as ``self``
            values to evaluate the pdf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            log pdf values at the given points
        """
        alpha = self._bc_params["alpha"]
        scale = self._bc_params["scale"]
        return np.log(alpha / x) + alpha * np.log(scale / x)

    def _cdf(self, x):
        """Cumulative distribution function.

        Parameters
        ----------
        x : 2D np.ndarray, same shape as ``self``
            values to evaluate the cdf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            cdf values at the given points
        """
        alpha = self._bc_params["alpha"]
        scale = self._bc_params["scale"]
        cdf_arr = np.where(x < scale, 0, 1 - np.power(scale / x, alpha))
        return cdf_arr

    def _ppf(self, p):
        """Quantile function = percent point function = inverse cdf.

        Parameters
        ----------
        p : 2D np.ndarray, same shape as ``self``
            values to evaluate the ppf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            ppf values at the given points
        """
        alpha = self._bc_params["alpha"]
        scale = self._bc_params["scale"]
        return scale / np.power(1 - p, 1 / alpha)

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator."""
        # array case examples
        params1 = {"scale": [[1, 1.5], [2, 3], [4, 5]], "alpha": 3}
        params2 = {
            "scale": 1,
            "alpha": 3,
            "index": pd.Index([1, 2, 5]),
            "columns": pd.Index(["a", "b"]),
        }
        # scalar case examples
        params3 = {"scale": 1, "alpha": 2}
        return [params1, params2, params3]



# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
"""Half-Logistic probability distribution."""

__author__ = ["SaiRevanth25"]

import pandas as pd
from scipy.stats import halflogistic, rv_continuous

from skpro.distributions.adapters.scipy import _ScipyAdapter


class HalfLogistic(_ScipyAdapter):
    r"""Half-Logistic distribution.

    Most methods wrap ``scipy.stats.halflogistic``.

    This distribution is univariate, without correlation between dimensions
    for the array-valued case.

    The half-logistic distribution is a continuous probability distribution derived
    from the logistic distribution by taking only the positive half. It is particularly
    useful in reliability analysis, lifetime modeling, and other applications where
    non-negative values are required.

    The half-logistic distribution is parametrized by the scale parameter
    :math:`\beta`, such that the pdf is

    .. math::

        f(x) = \frac{2 \exp\left(-\frac{x}{\beta}\right)}
                {\beta \left(1 + \exp\left(-\frac{x}{\beta}\right)\right)^2},
                x>0 otherwise 0

    The scale parameter :math:`\beta` is represented by the parameter ``beta``.

    Parameters
    ----------
    beta : float or array of float (1D or 2D), must be positive
        scale parameter of the half-logistic distribution
    index : pd.Index, optional, default = RangeIndex
    columns : pd.Index, optional, default = RangeIndex

    Examples
    --------
    >>> from skpro.distributions.halflogistic import HalfLogistic

    >>> hl = HalfLogistic(beta=1)
    """

    _tags = {
        "capabilities:approx": ["pdfnorm"],
        "capabilities:exact": ["mean", "var", "pdf", "log_pdf", "cdf", "ppf"],
        "distr:measuretype": "continuous",
        "distr:paramtype": "parametric",
        "broadcast_init": "on",
    }

    def __init__(self, beta, index=None, columns=None):
        self.beta = beta

        super().__init__(index=index, columns=columns)

    def _get_scipy_object(self) -> rv_continuous:
        return halflogistic

    def _get_scipy_param(self):
        beta = self._bc_params["beta"]
        return [beta], {}

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator."""
        # array case examples
        params1 = {"beta": [[1, 2], [3, 4]]}
        params2 = {
            "beta": 1,
            "index": pd.Index([1, 2, 5]),
            "columns": pd.Index(["a", "b"]),
        }
        # scalar case examples
        params3 = {"beta": 2}
        return [params1, params2, params3]



"""Doctest utilities."""
# copyright: sktime developers, BSD-3-Clause License (see LICENSE file)

import contextlib
import doctest
import io


def run_doctest(
    f,
    verbose=False,
    name=None,
    compileflags=None,
    optionflags=doctest.ELLIPSIS,
    raise_on_error=True,
):
    """Run doctests for a given function or class, and return or raise.

    Parameters
    ----------
    f : callable
        Function or class to run doctests for.
    verbose : bool, optional (default=False)
        If True, print the results of the doctests.
    name : str, optional (default=f.__name__, if available, otherwise "NoName")
        Name of the function or class.
    compileflags : int, optional (default=None)
        Flags to pass to the Python parser.
    optionflags : int, optional (default=doctest.ELLIPSIS)
        Flags to control the behaviour of the doctest.
    raise_on_error : bool, optional (default=True)
        If True, raise an exception if the doctests fail.

    Returns
    -------
    doctest_output : str
        Output of the doctests.

    Raises
    ------
    RuntimeError
        If raise_on_error=True and the doctests fail.
    """
    doctest_output_io = io.StringIO()
    with contextlib.redirect_stdout(doctest_output_io):
        doctest.run_docstring_examples(
            f=f,
            globs=globals(),
            verbose=verbose,
            name=name,
            compileflags=compileflags,
            optionflags=optionflags,
        )
    doctest_output = doctest_output_io.getvalue()

    if name is None:
        name = f.__name__ if hasattr(f, "__name__") else "NoName"

    if raise_on_error and len(doctest_output) > 0:
        raise RuntimeError(
            f"Docstring examples failed doctests "
            f"for {name}, doctest output: {doctest_output}"
        )
    return doctest_output
